from __future__ import annotations


__copyright__ = "Copyright (C) 2009-2013 Andreas Kloeckner"

__license__ = """
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""


from warnings import warn

from pymbolic.mapper import WalkMapper


class PersistentHashWalkMapper(WalkMapper):
    """A subclass of :class:`pymbolic.mapper.WalkMapper` for constructing
    persistent hash keys for use with
    :class:`pytools.persistent_dict.PersistentDict`.
    """

    def __init__(self, key_hash):
        self.key_hash = key_hash

        warn("PersistentHashWalkMapper is deprecated. "
             "Since they are dataclasses, expression objects should now "
             "support persistent hashing natively without any help. "
             "It will be removed in 2026.",
             DeprecationWarning, stacklevel=2)

    def visit(self, expr):
        self.key_hash.update(type(expr).__name__.encode("utf8"))
        return True

    def map_variable(self, expr):
        self.key_hash.update(expr.name.encode("utf8"))

    def map_constant(self, expr):
        import sys
        if "numpy" in sys.modules:
            import numpy as np
            if isinstance(expr, np.generic):
                # Makes a Python scalar from a numpy one.
                expr = expr.item()

        self.key_hash.update(repr(expr).encode("utf8"))

    def map_comparison(self, expr):
        if self.visit(expr):
            self.rec(expr.left)
            self.key_hash.update(repr(expr.operator).encode("utf8"))
            self.rec(expr.right)



"""
.. autoclass:: GraphvizMapper
"""
from __future__ import annotations


__copyright__ = "Copyright (C) 2015 Andreas Kloeckner"

__license__ = """
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""

from pymbolic.mapper import WalkMapper


class GraphvizMapper(WalkMapper):
    """Produces code for `dot <https://graphviz.org>`__ that yields
    an expression tree of the traversed expression(s).

    .. automethod:: get_dot_code

    .. versionadded:: 2015.1
    """

    def __init__(self):
        self.lines = []
        self.parent_stack = []

        self.next_unique_id = -1
        self.nodes_visited = set()
        self.common_subexpressions = {}

    def get_dot_code(self):
        """Return the dot source code for a previously traversed expression."""

        return "digraph expression {\n%s\n}" % (
            "\n".join("  "+line for line in self.lines))

    def get_id(self, expr):
        """Generate a unique node ID for dot for *expr*"""

        return f"id{id(expr)}"

    def map_leaf(self, expr):
        self.lines.append(
                '{} [label="{}", shape=box];'.format(
                    self.get_id(expr), str(expr).replace("\\", "\\\\")))

        if self.visit(expr, node_printed=True):
            self.post_visit(expr)

    def generate_unique_id(self):
        self.next_unique_id += 1
        return f"uid{self.next_unique_id}"

    def visit(self, expr, node_printed=False, node_id=None):
        # {{{ print connectivity

        if node_id is None:
            node_id = self.get_id(expr)

        if self.parent_stack:
            self.lines.append("{} -> {};".format(
                self.get_id(self.parent_stack[-1]),
                node_id))

        # }}}

        if id(expr) in self.nodes_visited:
            return False
        self.nodes_visited.add(id(expr))

        if not node_printed:
            self.lines.append(
                    '{} [label="{}"];'.format(
                        self.get_id(expr),
                        type(expr).__name__))

        self.parent_stack.append(expr)
        return True

    def post_visit(self, expr):
        self.parent_stack.pop(-1)

    def map_sum(self, expr):
        self.lines.append(
                '{} [label="+",shape=circle];'.format(self.get_id(expr)))
        if not self.visit(expr, node_printed=True):
            return

        for child in expr.children:
            self.rec(child)

        self.post_visit(expr)

    def map_product(self, expr):
        self.lines.append(
                '{} [label="*",shape=circle];'.format(self.get_id(expr)))
        if not self.visit(expr, node_printed=True):
            return

        for child in expr.children:
            self.rec(child)

        self.post_visit(expr)

    def map_variable(self, expr):
        # Shared nodes for variables do not make for pretty graphs.
        # So we generate our own unique IDs for them.

        node_id = self.generate_unique_id()

        self.lines.append(
                '{} [label="{}",shape=box];'.format(
                    node_id,
                    expr.name))
        if not self.visit(expr, node_printed=True, node_id=node_id):
            return

        self.post_visit(expr)

    def map_lookup(self, expr):
        self.lines.append(
                '{} [label="Lookup[{}]",shape=box];'.format(
                    self.get_id(expr), expr.name))
        if not self.visit(expr, node_printed=True):
            return

        self.rec(expr.aggregate)
        self.post_visit(expr)

    def map_constant(self, expr):
        # Some constants (Python ints notably) are shared among small (and
        # common) values. While accurate, this doesn't make for pretty
        # trees. So we generate our own unique IDs for them.

        node_id = self.generate_unique_id()

        self.lines.append(
                '{} [label="{}",shape=ellipse];'.format(
                    node_id,
                    str(expr)))
        if not self.visit(expr, node_printed=True, node_id=node_id):
            return

        self.post_visit(expr)

    def map_call(self, expr):
        from pymbolic.primitives import Variable
        if not isinstance(expr.function, Variable):
            return super().map_call(expr)

        self.lines.append(
                '{} [label="Call[{}]",shape=box];'.format(
                    self.get_id(expr), str(expr.function)))
        if not self.visit(expr, node_printed=True):
            return

        for child in expr.parameters:
            self.rec(child)

        self.post_visit(expr)

    def map_common_subexpression(self, expr):
        try:
            expr = self.common_subexpressions[expr]
        except KeyError:
            self.common_subexpressions[expr] = expr

        if not self.visit(expr):
            return

        self.rec(expr.child)

        self.post_visit(expr)

    # {{{ geometric algebra

    map_nabla_component = map_leaf
    map_nabla = map_leaf

    # }}}



from __future__ import annotations


__copyright__ = "Copyright (C) 2009-2013 Andreas Kloeckner"

__license__ = """
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""


from pymbolic.mapper import IdentityMapper, WalkMapper
from pymbolic.primitives import CommonSubexpression, cse_scope


class CSEWalkMapper(WalkMapper):
    def __init__(self):
        self.subexpr_histogram = {}

    def visit(self, expr):
        self.subexpr_histogram[expr] = self.subexpr_histogram.get(expr, 0) + 1
        return True


class CSETagMapper(IdentityMapper):
    def __init__(self, walk_mapper):
        self.subexpr_histogram = walk_mapper.subexpr_histogram

    def map_call(self, expr):
        if self.subexpr_histogram.get(expr, 0) > 1:
            return CommonSubexpression(expr, scope=cse_scope.EVALUATION)
        else:
            return getattr(IdentityMapper, expr.mapper_method)(self, expr)

    map_sum = map_call
    map_product = map_call
    map_quotient = map_call
    map_floor_div = map_call
    map_remainder = map_call
    map_power = map_call

    map_left_shift = map_call
    map_right_shift = map_call

    map_bitwise_not = map_call
    map_bitwise_or = map_call
    map_bitwise_xor = map_call
    map_bitwise_and = map_call

    map_comparison = map_call

    map_logical_not = map_call
    map_logical_and = map_call
    map_logical_or = map_call

    map_if = map_call
    map_if_positive = map_call



from __future__ import annotations

from pymbolic.typing import Expression


__copyright__ = "Copyright (C) 2009-2013 Andreas Kloeckner"

__license__ = """
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""

from collections.abc import Sequence
from sys import intern
from typing import ClassVar, TypeAlias

from immutabledict import immutabledict

import pytools.lex
from pytools import memoize_method


_imaginary = intern("imaginary")
_float = intern("float")
_int = intern("int")
_power = intern("exp")
_plus = intern("plus")
_minus = intern("minus")
_times = intern("times")
_floordiv = intern("floordiv")
_over = intern("over")
_modulo = intern("modulo")
_openpar = intern("openpar")
_closepar = intern("closepar")
_openbracket = intern("openbracket")
_closebracket = intern("closebracket")
_true = intern("True")
_false = intern("False")
_identifier = intern("identifier")
_whitespace = intern("whitespace")
_comma = intern("comma")
_dot = intern("dot")
_colon = intern("colon")

_assign = intern("assign")

_equal = intern("equal")
_notequal = intern("notequal")
_less = intern("less")
_lessequal = intern("lessequal")
_greater = intern("greater")
_greaterequal = intern("greaterequal")

_leftshift = intern("leftshift")
_rightshift = intern("rightshift")

_and = intern("and")
_or = intern("or")
_not = intern("not")
_if = intern("if")
_else = intern("else")

_bitwiseand = intern("bitwiseand")
_bitwiseor = intern("bitwiseor")
_bitwisexor = intern("bitwisexor")
_bitwisenot = intern("bitwisenot")

_PREC_COMMA = 5  # must be > 1 (1 is used by fortran-to-cl)
_PREC_SLICE = 10
_PREC_IF = 75
_PREC_LOGICAL_OR = 80
_PREC_LOGICAL_AND = 90

_PREC_BITWISE_OR = 120
_PREC_BITWISE_XOR = 120
_PREC_BITWISE_AND = 130

_PREC_COMPARISON = 200
_PREC_SHIFT = 205
_PREC_PLUS = 210
_PREC_TIMES = 220
_PREC_UNARY = 230
_PREC_POWER = 240
_PREC_CALL = 250


def _join_to_slice(left, right):
    from pymbolic.primitives import Slice
    if isinstance(right, Slice):
        return Slice((left, *right.children))
    else:
        return Slice((left, right))


class FinalizedContainer:
    """A base class for containers  that may not have elements appended to it,
    because they were terminated by a closing delimiter.
    """


class FinalizedTuple(tuple, FinalizedContainer):
    pass


class FinalizedList(list, FinalizedContainer):
    @memoize_method
    def __hash__(self) -> int:  # type: ignore[override]
        result = hash(type(self).__name__)
        for it in self:
            result ^= hash(it)

        return result


LexTable: TypeAlias = Sequence[
        tuple[str, pytools.lex.RE | tuple[str | pytools.lex.RE, ...]]]


class Parser:
    lex_table: ClassVar[LexTable] = [
            (_equal, pytools.lex.RE(r"==")),
            (_notequal, pytools.lex.RE(r"!=")),
            (_equal, pytools.lex.RE(r"==")),

            (_leftshift, pytools.lex.RE(r"\<\<")),
            (_rightshift, pytools.lex.RE(r"\>\>")),

            (_lessequal, pytools.lex.RE(r"\<=")),
            (_greaterequal, pytools.lex.RE(r"\>=")),
            # must be before
            (_less, pytools.lex.RE(r"\<")),
            (_greater, pytools.lex.RE(r"\>")),

            (_assign, pytools.lex.RE(r"=")),

            (_and, pytools.lex.RE(r"and\b")),
            (_or, pytools.lex.RE(r"or\b")),
            (_not, pytools.lex.RE(r"not\b")),
            (_if, pytools.lex.RE(r"if\b")),
            (_else, pytools.lex.RE(r"else\b")),

            (_imaginary, (_float, pytools.lex.RE("j"))),
            (_float, ("|",
                # has digits before the dot (after optional)
                pytools.lex.RE(
                    r"[0-9]+\.[0-9]*([eEdD][+-]?[0-9]+)?([a-zA-Z]*)"),
                pytools.lex.RE(
                    r"[0-9]+(\.[0-9]*)?[eEdD][+-]?[0-9]+([a-zA-Z]*)\b"),
                # has digits after the dot (before optional)
                pytools.lex.RE(
                    r"[0-9]*\.[0-9]+([eEdD][+-]?[0-9]+)?([a-zA-Z]*)"),
                pytools.lex.RE(
                    r"[0-9]*\.[0-9]+[eEdD][+-]?[0-9]+([a-zA-Z]*)\b"),
                # has a letter tag
                pytools.lex.RE(r"[0-9]+([a-zA-Z]+)"),
                )),
            (_int, pytools.lex.RE(r"[0-9]+")),

            (_plus, pytools.lex.RE(r"\+")),
            (_minus, pytools.lex.RE(r"-")),
            (_power, pytools.lex.RE(r"\*\*")),
            (_times, pytools.lex.RE(r"\*")),
            (_floordiv, pytools.lex.RE(r"//")),
            (_over, pytools.lex.RE(r"/")),
            (_modulo, pytools.lex.RE(r"%")),

            (_bitwiseand, pytools.lex.RE(r"\&")),
            (_bitwiseor, pytools.lex.RE(r"\|")),
            (_bitwisenot, pytools.lex.RE(r"\~")),
            (_bitwisexor, pytools.lex.RE(r"\^")),

            (_openpar, pytools.lex.RE(r"\(")),
            (_closepar, pytools.lex.RE(r"\)")),
            (_openbracket, pytools.lex.RE(r"\[")),
            (_closebracket, pytools.lex.RE(r"\]")),
            (_true, pytools.lex.RE(r"True")),
            (_false, pytools.lex.RE(r"False")),
            (_identifier, pytools.lex.RE(r"[@$a-z_A-Z_][@$a-zA-Z_0-9]*")),
            (_whitespace, pytools.lex.RE("[ \n\t]*")),
            (_comma, pytools.lex.RE(",")),
            (_dot, pytools.lex.RE(r"\.")),
            (_colon, pytools.lex.RE(r"\:")),
            ]

    _COMP_TABLE: ClassVar[dict[str, str]] = {
            _greater: ">",
            _greaterequal: ">=",
            _less: "<",
            _lessequal: "<=",
            _equal: "==",
            _notequal: "!=",
            }

    def parse_float(self, s):
        return float(s.replace("d", "e").replace("D", "e"))

    def parse_terminal(self, pstate):
        import pymbolic.primitives as primitives

        next_tag = pstate.next_tag()
        if next_tag is _int:
            return int(pstate.next_str_and_advance())
        elif next_tag is _float:
            return self.parse_float(pstate.next_str_and_advance())
        elif next_tag is _imaginary:
            return complex(pstate.next_str_and_advance())
        elif next_tag is _true:
            assert pstate.next_str_and_advance() == "True"
            return True
        elif next_tag is _false:
            assert pstate.next_str_and_advance() == "False"
            return False
        elif next_tag is _identifier:
            return primitives.Variable(pstate.next_str_and_advance())
        elif next_tag is _if:
            from warnings import warn
            warn("Usage of 'if' as an identifier is deprecated due to"
                    " introduction of python style 'if-else' expressions.",
                    DeprecationWarning, stacklevel=2)
            return primitives.Variable(pstate.next_str_and_advance())
        else:
            pstate.expected("terminal")

    def parse_prefix(self, pstate):
        import pymbolic.primitives as primitives
        pstate.expect_not_end()

        if pstate.is_next(_colon):
            pstate.advance()

            expr_pstate = pstate.copy()
            from pytools.lex import ParseError
            try:
                next_expr = self.parse_expression(expr_pstate, _PREC_SLICE)
            except ParseError:
                # no expression follows, too bad.
                left_exp = primitives.Slice((None,))
            else:
                left_exp = _join_to_slice(None, next_expr)
                pstate.assign(expr_pstate)
        elif pstate.is_next(_times):
            pstate.advance()
            left_exp = primitives.Wildcard()
        elif pstate.is_next(_plus):
            pstate.advance()
            left_exp = self.parse_expression(pstate, _PREC_UNARY)
        elif pstate.is_next(_minus):
            pstate.advance()
            left_exp = -self.parse_expression(pstate, _PREC_UNARY)  # pylint:disable=invalid-unary-operand-type
        elif pstate.is_next(_not):
            pstate.advance()
            from pymbolic.primitives import LogicalNot
            left_exp = LogicalNot(
                    self.parse_expression(pstate, _PREC_UNARY))
        elif pstate.is_next(_bitwisenot):
            pstate.advance()
            from pymbolic.primitives import BitwiseNot
            left_exp = BitwiseNot(
                    self.parse_expression(pstate, _PREC_UNARY))
        elif pstate.is_next(_openpar):
            pstate.advance()

            if pstate.is_next(_closepar):
                left_exp = ()
            else:
                # This is parsing expressions separated by commas, so it
                # will return a tuple. Kind of the lazy way out.
                left_exp = self.parse_expression(pstate)

            pstate.expect(_closepar)
            pstate.advance()
            if isinstance(left_exp, tuple):
                # These could just be plain parentheses.

                # Finalization prevents things from being appended
                # to containers after their closing delimiter.
                left_exp = FinalizedTuple(left_exp)
        elif pstate.is_next(_openbracket):
            pstate.advance()

            if pstate.is_next(_closebracket):
                left_exp = ()
            else:
                # This is parsing expressions separated by commas, so it
                # will return a tuple. Kind of the lazy way out.
                left_exp = self.parse_expression(pstate)

            pstate.expect(_closebracket)
            pstate.advance()

            # Finalization prevents things from being appended
            # to containers after their closing delimiter.
            if isinstance(left_exp, tuple):
                left_exp = FinalizedList(left_exp)
            else:
                left_exp = FinalizedList([left_exp])

        else:
            left_exp = self.parse_terminal(pstate)

        return left_exp

    def parse_expression(self, pstate, min_precedence=0):
        left_exp = self.parse_prefix(pstate)

        did_something = True
        while did_something:
            did_something = False
            if pstate.is_at_end():
                return left_exp

            result = self.parse_postfix(
                    pstate, min_precedence, left_exp)
            left_exp, did_something = result

        if isinstance(left_exp, FinalizedTuple):
            return tuple(left_exp)

        return left_exp

    def parse_postfix(self, pstate, min_precedence, left_exp):
        import pymbolic.primitives as primitives

        did_something = False

        next_tag = pstate.next_tag()

        if next_tag is _openpar and _PREC_CALL > min_precedence:
            pstate.advance()
            args, kwargs = self.parse_arglist(pstate)

            if kwargs:
                left_exp = primitives.CallWithKwargs(
                        left_exp, args, immutabledict(kwargs))
            else:
                left_exp = primitives.Call(left_exp, args)

            did_something = True
        elif next_tag is _openbracket and _PREC_CALL > min_precedence:
            pstate.advance()
            pstate.expect_not_end()
            left_exp = primitives.Subscript(left_exp, self.parse_expression(pstate))
            pstate.expect(_closebracket)
            pstate.advance()
            did_something = True
        elif next_tag is _if and _PREC_IF > min_precedence:
            from pymbolic.primitives import If
            then_expr = left_exp
            pstate.advance()
            pstate.expect_not_end()
            condition = self.parse_expression(pstate, _PREC_IF)
            pstate.expect(_else)
            pstate.advance()
            else_expr = self.parse_expression(pstate)
            left_exp = If(condition, then_expr, else_expr)
            did_something = True
        elif next_tag is _dot and _PREC_CALL > min_precedence:
            pstate.advance()
            pstate.expect(_identifier)
            left_exp = primitives.Lookup(left_exp, pstate.next_str())
            pstate.advance()
            did_something = True
        elif next_tag is _plus and _PREC_PLUS > min_precedence:
            pstate.advance()
            right_exp = self.parse_expression(pstate, _PREC_PLUS)
            if isinstance(left_exp, primitives.Sum):
                left_exp = primitives.Sum((*left_exp.children, right_exp))
            else:
                left_exp = primitives.Sum((left_exp, right_exp))

            did_something = True
        elif next_tag is _minus and _PREC_PLUS > min_precedence:
            pstate.advance()
            right_exp = self.parse_expression(pstate, _PREC_PLUS)
            if isinstance(left_exp, primitives.Sum):
                left_exp = primitives.Sum(left_exp.children + ((-right_exp),))  # noqa pylint:disable=invalid-unary-operand-type
            else:
                left_exp = primitives.Sum((left_exp, -right_exp))  # pylint:disable=invalid-unary-operand-type
            did_something = True
        elif next_tag is _times and _PREC_TIMES > min_precedence:
            pstate.advance()
            right_exp = self.parse_expression(pstate, _PREC_PLUS)
            if isinstance(left_exp, primitives.Product):
                left_exp = primitives.Product((*left_exp.children, right_exp))
            else:
                left_exp = primitives.Product((left_exp, right_exp))
            did_something = True
        elif next_tag is _floordiv and _PREC_TIMES > min_precedence:
            pstate.advance()
            left_exp = primitives.FloorDiv(
                    left_exp, self.parse_expression(pstate, _PREC_TIMES))
            did_something = True
        elif next_tag is _over and _PREC_TIMES > min_precedence:
            pstate.advance()
            left_exp = primitives.Quotient(
                    left_exp, self.parse_expression(pstate, _PREC_TIMES))
            did_something = True
        elif next_tag is _modulo and _PREC_TIMES > min_precedence:
            pstate.advance()
            left_exp = primitives.Remainder(
                    left_exp, self.parse_expression(pstate, _PREC_TIMES))
            did_something = True
        elif next_tag is _power and _PREC_POWER > min_precedence:
            pstate.advance()
            left_exp = primitives.Power(
                    left_exp, self.parse_expression(pstate, _PREC_TIMES))
            did_something = True
        elif next_tag is _and and _PREC_LOGICAL_AND > min_precedence:
            pstate.advance()
            from pymbolic.primitives import LogicalAnd
            left_exp = LogicalAnd((
                    left_exp,
                    self.parse_expression(pstate, _PREC_LOGICAL_AND)))
            did_something = True
        elif next_tag is _or and _PREC_LOGICAL_OR > min_precedence:
            pstate.advance()
            from pymbolic.primitives import LogicalOr
            left_exp = LogicalOr((
                    left_exp,
                    self.parse_expression(pstate, _PREC_LOGICAL_OR)))
            did_something = True
        elif next_tag is _bitwiseor and _PREC_BITWISE_OR > min_precedence:
            pstate.advance()
            from pymbolic.primitives import BitwiseOr
            left_exp = BitwiseOr((
                    left_exp,
                    self.parse_expression(pstate, _PREC_BITWISE_OR)))
            did_something = True
        elif next_tag is _bitwisexor and _PREC_BITWISE_XOR > min_precedence:
            pstate.advance()
            from pymbolic.primitives import BitwiseXor
            left_exp = BitwiseXor((
                    left_exp,
                    self.parse_expression(pstate, _PREC_BITWISE_XOR)))
            did_something = True
        elif next_tag is _bitwiseand and _PREC_BITWISE_AND > min_precedence:
            pstate.advance()
            from pymbolic.primitives import BitwiseAnd
            left_exp = BitwiseAnd((
                    left_exp,
                    self.parse_expression(pstate, _PREC_BITWISE_AND)))
            did_something = True
        elif next_tag is _rightshift and _PREC_SHIFT > min_precedence:
            pstate.advance()
            from pymbolic.primitives import RightShift
            left_exp = RightShift(
                    left_exp,
                    self.parse_expression(pstate, _PREC_SHIFT))
            did_something = True
        elif next_tag is _leftshift and _PREC_SHIFT > min_precedence:
            pstate.advance()
            from pymbolic.primitives import LeftShift
            left_exp = LeftShift(
                    left_exp,
                    self.parse_expression(pstate, _PREC_SHIFT))
            did_something = True
        elif next_tag in self._COMP_TABLE and _PREC_COMPARISON > min_precedence:
            pstate.advance()
            from pymbolic.primitives import Comparison
            left_exp = Comparison(
                    left_exp,
                    self._COMP_TABLE[next_tag],
                    self.parse_expression(pstate, _PREC_COMPARISON))
            did_something = True
        elif next_tag is _colon and _PREC_SLICE >= min_precedence:
            pstate.advance()
            expr_pstate = pstate.copy()

            assert not isinstance(left_exp, primitives.Slice)

            from pytools.lex import ParseError
            try:
                next_expr = self.parse_expression(expr_pstate, _PREC_SLICE)
            except ParseError:
                # no expression follows, too bad.
                left_exp = primitives.Slice((left_exp, None,))
            else:
                left_exp = _join_to_slice(left_exp, next_expr)
                pstate.assign(expr_pstate)

            did_something = True

        elif next_tag is _comma and _PREC_COMMA > min_precedence:
            # The precedence makes the comma left-associative.

            pstate.advance()
            if pstate.is_at_end() or pstate.next_tag() is _closepar:
                if isinstance(left_exp, tuple | list) \
                        and not isinstance(left_exp, FinalizedContainer):
                    # left_expr is a container with trailing commas
                    pass
                else:
                    left_exp = (left_exp,)
            else:
                new_el = self.parse_expression(pstate, _PREC_COMMA)
                if isinstance(left_exp, tuple | list) \
                        and not isinstance(left_exp, FinalizedContainer):
                    left_exp = (*left_exp, new_el)
                else:
                    left_exp = (left_exp, new_el)

            did_something = True

        return left_exp, did_something

    def parse_arglist(self, pstate):
        pstate.expect_not_end()

        args = []
        kwargs = {}

        comma_allowed = False
        while True:
            pstate.expect_not_end()

            saw_comma = False
            if pstate.next_tag() is _comma:
                saw_comma = True
                if not comma_allowed:
                    pstate.raise_parse_error("comma not expected")
                pstate.advance()
                pstate.expect_not_end()

            if pstate.next_tag() is _closepar:
                pstate.advance()
                return tuple(args), kwargs

            if not saw_comma and comma_allowed:
                pstate.raise_parse_error("comma expected")

            if (pstate.next_tag() is _identifier
                    and not pstate.is_at_end(1)
                    and pstate.next_tag(1) == _assign):
                kw = pstate.next_str()
                pstate.advance()
                pstate.advance()

                kwargs[kw] = self.parse_expression(pstate, _PREC_COMMA)
            else:
                if kwargs:
                    pstate.raise_parse_error(
                            "positional argument after keyword "
                            "argument not allowed")

                args.append(self.parse_expression(pstate, _PREC_COMMA))

            comma_allowed = True

    def __call__(self, expr_str: str, min_precedence: int = 0) -> Expression:
        lex_result = [(tag, s, idx, match_obj)
                for (tag, s, idx, match_obj) in pytools.lex.lex(
                    self.lex_table, expr_str,
                    match_objects=True)
                if tag is not _whitespace]
        pstate = pytools.lex.LexIterator(lex_result, expr_str)

        result = self. parse_expression(pstate, min_precedence)
        if not pstate.is_at_end():
            pstate.raise_parse_error("leftover input after completed parse")
        return result


parse = Parser()



from __future__ import annotations


__copyright__ = "Copyright (C) 2009-2013 Andreas Kloeckner"

__license__ = """
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""

import re
from collections.abc import Callable, Iterable, Mapping
from dataclasses import dataclass, fields
from functools import partial
from sys import intern
from typing import (
    TYPE_CHECKING,
    Any,
    ClassVar,
    NoReturn,
    Protocol,
    TypeAlias,
    TypeVar,
    cast,
)
from warnings import warn

from immutabledict import immutabledict
from typing_extensions import TypeIs, dataclass_transform

from pytools import module_getattr_for_deprecations

from . import traits
from .typing import ArithmeticExpression, Expression as _Expression, Number, Scalar


if TYPE_CHECKING:
    from _typeshed import DataclassInstance


__doc__ = """
Expression base class
---------------------

.. currentmodule:: pymbolic

.. autoclass:: ExpressionNode

.. autofunction:: expr_dataclass

.. autoclass:: Variable
    :show-inheritance:
    :undoc-members:
    :members: mapper_method

Expressions
-----------

.. currentmodule:: pymbolic.primitives

.. autoclass:: AlgebraicLeaf
    :show-inheritance:
    :undoc-members:
    :members: mapper_method

.. autoclass:: Leaf
    :show-inheritance:
    :undoc-members:
    :members: mapper_method

.. autoclass:: Wildcard
    :show-inheritance:
    :undoc-members:
    :members: mapper_method

.. autoclass:: DotWildcard
    :show-inheritance:
    :undoc-members:
    :members: mapper_method

.. autoclass:: StarWildcard
    :show-inheritance:
    :undoc-members:
    :members: mapper_method

Function Calls
--------------

.. autoclass:: Call
    :show-inheritance:
    :undoc-members:
    :members: mapper_method

.. autoclass:: CallWithKwargs
    :show-inheritance:
    :undoc-members:
    :members: mapper_method

.. autoclass:: Subscript
    :show-inheritance:
    :undoc-members:
    :members: mapper_method

.. autoclass:: Lookup
    :show-inheritance:
    :undoc-members:
    :members: mapper_method

Sums, products and such
-----------------------

.. autoclass:: Sum
    :show-inheritance:
    :undoc-members:
    :members: mapper_method

.. autoclass:: Product
    :show-inheritance:
    :undoc-members:
    :members: mapper_method

.. autoclass:: Min
    :show-inheritance:
    :undoc-members:
    :members: mapper_method

.. autoclass:: Max
    :show-inheritance:
    :undoc-members:
    :members: mapper_method

.. autoclass:: Quotient
    :undoc-members:
    :members: mapper_method

.. autoclass:: FloorDiv
    :undoc-members:
    :members: mapper_method

.. autoclass:: Remainder
    :undoc-members:
    :members: mapper_method

.. autoclass:: Power
    :show-inheritance:
    :undoc-members:
    :members: mapper_method

Shift operators
---------------

.. autoclass:: LeftShift
    :undoc-members:
    :members: mapper_method

.. autoclass:: RightShift
    :undoc-members:
    :members: mapper_method

Bitwise operators
-----------------

.. autoclass:: BitwiseNot
    :show-inheritance:
    :undoc-members:
    :members: mapper_method

.. autoclass:: BitwiseOr
    :show-inheritance:
    :undoc-members:
    :members: mapper_method

.. autoclass:: BitwiseXor
    :show-inheritance:
    :undoc-members:
    :members: mapper_method

.. autoclass:: BitwiseAnd
    :show-inheritance:
    :undoc-members:
    :members: mapper_method

Comparisons and logic
---------------------

.. autoclass:: Comparison
    :show-inheritance:
    :undoc-members:
    :members: mapper_method

.. autoclass:: LogicalNot
    :show-inheritance:
    :undoc-members:
    :members: mapper_method

.. autoclass:: LogicalAnd
    :show-inheritance:
    :undoc-members:
    :members: mapper_method

.. autoclass:: LogicalOr
    :show-inheritance:
    :undoc-members:
    :members: mapper_method

.. autoclass:: If
    :show-inheritance:
    :undoc-members:
    :members: mapper_method

Slices
----------

.. autoclass:: Slice
    :show-inheritance:
    :undoc-members:
    :members: mapper_method

Code generation helpers
-----------------------

.. autoclass:: CommonSubexpression
    :show-inheritance:
    :undoc-members:
    :members: mapper_method

.. autoclass:: cse_scope
.. autofunction:: make_common_subexpression

Symbolic derivatives and substitution
-------------------------------------

Inspired by similar functionality in :mod:`sympy`.

.. autoclass:: Substitution
    :show-inheritance:
    :undoc-members:
    :members: mapper_method

.. autoclass:: Derivative
    :show-inheritance:
    :undoc-members:
    :members: mapper_method

Helper functions
----------------

.. autofunction:: is_zero
.. autofunction:: is_constant
.. autofunction:: flattened_sum
.. autofunction:: flattened_product
.. autofunction:: register_constant_class
.. autofunction:: unregister_constant_class
.. autofunction:: variables

Interaction with :mod:`numpy` arrays
------------------------------------

:class:`numpy.ndarray` instances are supported anywhere in an expression.
In particular, :mod:`numpy` object arrays are useful for capturing
vectors and matrices of :mod:`pymbolic` objects.

.. autofunction:: make_sym_vector
.. autofunction:: make_sym_array

Constants
---------

.. autoclass:: NaN
    :show-inheritance:
    :undoc-members:
    :members: mapper_method

Helper classes
--------------

.. autoclass:: EmptyOK
.. autoclass:: _AttributeLookupCreator

References
----------

.. class:: DataclassInstance

    An instance of a :func:`~dataclasses.dataclass`.

.. class:: ArithmeticExpression

    See :class:`pymbolic.ArithmeticExpression`

.. class:: _T

    A type variable.

.. currentmodule:: numpy

.. class:: bool_

    Deprecated alias for :class:`numpy.bool`.

.. currentmodule:: typing_extensions

.. class:: TypeIs

    See :data:`typing_extensions.TypeIs`.

.. class:: Variable

    See :class:`pymbolic.Variable`.

.. class:: Expression

    See :data:`pymbolic.typing.Expression`.

.. currentmodule:: pymbolic

.. class:: Comparison

    See :class:`pymbolic.primitives.Comparison`.

.. class:: LogicalNot

    See :class:`pymbolic.primitives.LogicalNot`.

.. class:: LogicalAnd

    See :class:`pymbolic.primitives.LogicalAnd`.

.. class:: LogicalOr

    See :class:`pymbolic.primitives.LogicalOr`.

.. class:: Lookup

    See :class:`pymbolic.primitives.Lookup`.
"""


# The upper bound for the number of nodes printed when repr
# is called on a pymbolic object.
SAFE_REPR_LIMIT = 10


def disable_subscript_by_getitem():
    # The issue that was addressed by this could be fixed
    # in a much less ham-fisted manner, and thus this has been
    # made a no-op.
    #
    # See
    # https://github.com/inducer/pymbolic/issues/4
    pass


# https://stackoverflow.com/a/13624858
class _classproperty(property):  # noqa: N801
    def __get__(self, owner_self, owner_cls):
        return self.fget(owner_cls)


class _AttributeLookupCreator:
    """Helper used by :attr:`pymbolic.ExpressionNode.a` to create lookups.

    .. automethod:: __getattr__
    """
    def __init__(self, aggregate: _Expression) -> None:
        self.aggregate = aggregate

    def __getattr__(self, name: str) -> Lookup:
        return Lookup(self.aggregate, name)


@dataclass(frozen=True)
class EmptyOK:
    child: _Expression


class ExpressionNode:
    """Superclass for parts of a mathematical expression. Overrides operators
    to implicitly construct :class:`~pymbolic.primitives.Sum`,
    :class:`~pymbolic.primitives.Product` and other expressions.

    Expression objects are immutable.

    .. versionchanged:: 2022.2

        `PEP 634 <https://peps.python.org/pep-0634/>`__-style pattern matching
        is now supported when Pymbolic is used under Python 3.10.

    .. autoproperty:: a

    .. automethod:: attr

    .. autoattribute:: mapper_method

        The :class:`pymbolic.mapper.Mapper` method called for objects of
        this type.

    .. automethod:: __getitem__

    .. automethod:: make_stringifier

    .. automethod:: __eq__
    .. automethod:: __hash__
    .. automethod:: __str__
    .. automethod:: __repr__

    .. rubric:: Logical operator constructors

    .. automethod:: not_
    .. automethod:: and_
    .. automethod:: or_

    .. rubric:: Comparison constructors

    .. automethod:: eq
    .. automethod:: ne
    .. automethod:: lt
    .. automethod:: le
    .. automethod:: gt
    .. automethod:: ge
    """

    mapper_method: ClassVar[str]

    # {{{ init arg names (override by subclass)

    def __getinitargs__(self):
        raise NotImplementedError

    @_classproperty
    def __match_args__(cls):  # noqa: N805  # pylint: disable=no-self-argument
        return cls.init_arg_names

    @property
    def init_arg_names(self) -> tuple[str, ...]:
        raise NotImplementedError

    # }}}

    # {{{ arithmetic

    def __add__(self, other: object) -> Sum:
        if not is_arithmetic_expression(other):
            return NotImplemented
        return Sum((self, other))

    def __radd__(self, other: object) -> Sum:
        if not is_arithmetic_expression(other):
            return NotImplemented
        return Sum((other, self))

    def __sub__(self, other: object) -> Sum:
        if not is_arithmetic_expression(other):
            return NotImplemented
        return Sum((self, -other))

    def __rsub__(self, other: object) -> Sum:
        if not is_arithmetic_expression(other):
            return NotImplemented
        return Sum((other, -self))

    def __mul__(self, other: object) -> Product:
        if not is_valid_operand(other):
            return NotImplemented

        return Product((self, other))

    def __rmul__(self, other: object) -> Product:
        if not is_valid_operand(other):
            return NotImplemented

        return Product((other, self))

    def __truediv__(self, other: object) -> Quotient:
        if not is_arithmetic_expression(other):
            return NotImplemented

        return Quotient(self, other)

    def __rtruediv__(self, other: object) -> Quotient:
        if not is_arithmetic_expression(other):
            return NotImplemented

        return Quotient(other, self)

    def __floordiv__(self, other: object) -> FloorDiv:
        if not is_arithmetic_expression(other):
            return NotImplemented

        return FloorDiv(self, other)

    def __rfloordiv__(self, other: object) -> FloorDiv:
        if not is_arithmetic_expression(other):
            return NotImplemented

        return FloorDiv(other, self)

    def __mod__(self, other: object) -> Remainder:
        if not is_arithmetic_expression(other):
            return NotImplemented

        return Remainder(self, other)

    def __rmod__(self, other: object) -> Remainder:
        if not is_arithmetic_expression(other):
            return NotImplemented

        return Remainder(other, self)

    def __pow__(self, other: object) -> Power:
        if not is_arithmetic_expression(other):
            return NotImplemented

        return Power(self, other)

    def __rpow__(self, other: object) -> Power:
        if not is_arithmetic_expression(other):
            return NotImplemented

        return Power(other, self)

    # }}}

    # {{{ shifts

    def __lshift__(self, other: object) -> LeftShift:
        if not is_valid_operand(other):
            return NotImplemented

        return LeftShift(self, other)

    def __rlshift__(self, other: object) -> LeftShift:
        if not is_valid_operand(other):
            return NotImplemented

        return LeftShift(other, self)

    def __rshift__(self, other: object) -> RightShift:
        if not is_valid_operand(other):
            return NotImplemented

        return RightShift(self, other)

    def __rrshift__(self, other: object) -> RightShift:
        if not is_valid_operand(other):
            return NotImplemented

        return RightShift(other, self)

    # }}}

    # {{{ bitwise operators

    def __invert__(self) -> BitwiseNot:
        return BitwiseNot(self)

    def __or__(self, other: object) -> BitwiseOr:
        if not is_valid_operand(other):
            return NotImplemented

        return BitwiseOr((self, other))

    def __ror__(self, other: object) -> BitwiseOr:
        if not is_valid_operand(other):
            return NotImplemented

        return BitwiseOr((other, self))

    def __xor__(self, other: object) -> BitwiseXor:
        if not is_valid_operand(other):
            return NotImplemented

        return BitwiseXor((self, other))

    def __rxor__(self, other: object) -> BitwiseXor:
        if not is_valid_operand(other):
            return NotImplemented

        return BitwiseXor((other, self))

    def __and__(self, other: object) -> BitwiseAnd:
        if not is_valid_operand(other):
            return NotImplemented

        return BitwiseAnd((self, other))

    def __rand__(self, other: object) -> BitwiseAnd:
        if not is_valid_operand(other):
            return NotImplemented

        return BitwiseAnd((other, self))

    # }}}

    # {{{ misc

    def __neg__(self) -> ArithmeticExpression:
        return -1*self

    def __pos__(self) -> ArithmeticExpression:
        return self

    def __call__(self, *args, **kwargs) -> Call | CallWithKwargs:
        if kwargs:
            from immutabledict import immutabledict
            return CallWithKwargs(self, args, immutabledict(kwargs))
        else:
            return Call(self, args)

    if not TYPE_CHECKING:
        # Subscript has an attribute 'index' which can't coexist with this.
        # Thus we're hiding this from mypy until it goes away.

        def index(self, subscript: ExpressionNode) -> ExpressionNode:
            """Return an expression representing ``self[subscript]``.

            .. versionadded:: 2014.3
            """

            warn("Expression.index(i) is deprecated and will be removed in 2H2025. "
                 "Use 'expr[i] 'instead.", DeprecationWarning, stacklevel=2)

            return self[subscript]

    def __getitem__(self, subscript: _Expression | EmptyOK) -> ExpressionNode:
        """Return an expression representing ``self[subscript]``. """

        if isinstance(subscript, EmptyOK):
            return Subscript(self, subscript.child)

        if subscript == ():
            warn("Expression.__getitem__ called with an empty tuple as an index. "
                 "This still returns just the aggregate (not a Subscript), "
                 "but this behavior will change in 2026. To avoid this warning "
                 "(and return a Subscript unconditionally), wrap the subscript "
                 "in pymbolic.primitives.EmptyOK.", DeprecationWarning, stacklevel=2)
            return self
        return Subscript(self, subscript)

    def attr(self, name: str) -> Lookup:
        """Return a :class:`Lookup` for *name* in *self*.
        """
        return Lookup(self, name)

    @property
    def a(self) -> _AttributeLookupCreator:
        """Provide a spelling ``expr.a.name`` for encoding attribute lookup.
        """
        return _AttributeLookupCreator(self)

    def __float__(self) -> float:
        from pymbolic.mapper.evaluator import evaluate_to_float
        return evaluate_to_float(self)

    def make_stringifier(self, originating_stringifier=None):
        """Return a :class:`pymbolic.mapper.Mapper` instance that can
        be used to generate a human-readable representation of *self*. Usually
        a subclass of :class:`pymbolic.mapper.stringifier.StringifyMapper`.

        :arg originating_stringifier: If provided, the newly created
            stringifier should carry forward attributes and settings of
            *originating_stringifier*.
        """
        from pymbolic.mapper.stringifier import StringifyMapper
        return StringifyMapper()

    def __str__(self) -> str:
        """Use the :meth:`make_stringifier` to return a human-readable
        string representation of *self*.
        """

        from pymbolic.mapper.stringifier import PREC_NONE
        return self.make_stringifier()(self, PREC_NONE)

    def _safe_repr(self, limit: int | None = None) -> str:
        if limit is None:
            limit = SAFE_REPR_LIMIT

        if limit <= 0:
            return "..."

        def strify_child(child, limit):
            if isinstance(child, tuple):
                # Make sure limit propagates at least through tuples

                return "({}{})".format(
                        ", ".join(strify_child(i, limit-1) for i in child),
                        "," if len(child) == 1 else "")

            elif isinstance(child, ExpressionNode):
                return child._safe_repr(limit=limit-1)
            else:
                return repr(child)

        from dataclasses import is_dataclass

        # NOTE: __getinitargs__() is deprecated, so skip it if possible
        if is_dataclass(self):
            initargs = tuple(getattr(self, fld.name) for fld in fields(self))
        else:
            initargs = self.__getinitargs__()
        initargs_str = ", ".join(strify_child(i, limit - 1) for i in initargs)

        return f"{self.__class__.__name__}({initargs_str})"

    def __repr__(self) -> str:
        return self._safe_repr()

    # }}}

    # {{{ hash/equality interface

    # This custom warning deduplication mechanism became necessary because the
    # sheer amount of warnings ended up leading to out-of-memory situations
    # with pytest which bufered all the warnings.
    _deprecation_warnings_issued: ClassVar[set[tuple[type[ExpressionNode], str]]] \
        = set()

    def __eq__(self, other) -> bool:
        """Provides equality testing with quick positive and negative paths
        based on :func:`id` and :meth:`__hash__`.
        """
        depr_key = (type(self), "__eq__")
        if depr_key not in self._deprecation_warnings_issued:
            warn(f"Expression.__eq__ is used by {self.__class__}. This is deprecated. "
                 "Use equality comparison supplied by expr_dataclass instead. "
                 "This will stop working in 2025.",
                 DeprecationWarning, stacklevel=2)
            self._deprecation_warnings_issued.add(depr_key)

        if self is other:
            return True
        elif hash(self) != hash(other):
            return False
        else:
            return self.is_equal(other)

    def __ne__(self, other) -> bool:
        return not self.__eq__(other)

    def __hash__(self) -> int:
        """Provides caching for hash values.
        """
        depr_key = (type(self), "__hash__")
        if depr_key not in self._deprecation_warnings_issued:
            warn(f"Expression.__hash__ is used by {self.__class__}. "
                 "This is deprecated. "
                 "Use hash functions supplied by expr_dataclass instead. "
                 "This will stop working in 2025.",
                 DeprecationWarning, stacklevel=2)

            self._deprecation_warnings_issued.add(depr_key)

        try:
            return self._hash_value
        except AttributeError:
            self._hash_value: int = self.get_hash()
            return self._hash_value

    def __getstate__(self) -> tuple[Any]:
        return self.__getinitargs__()

    def __setstate__(self, state) -> None:
        # Can't use trivial pickling: _hash_value cache must stay unset
        assert len(self.init_arg_names) == len(state), type(self)
        for name, value in zip(self.init_arg_names, state, strict=True):
            object.__setattr__(self, name, value)

    # }}}

    # {{{ hash/equality backend

    def is_equal(self, other) -> bool:
        return (type(other) is type(self)
                and self.__getinitargs__() == other.__getinitargs__())

    def get_hash(self) -> int:
        return hash((type(self).__name__, *self.__getinitargs__()))

    # }}}

    # {{{ logical op constructors

    def not_(self) -> LogicalNot:
        """Return *self* wrapped in a :class:`LogicalNot`.

        .. versionadded:: 2015.2
        """
        return LogicalNot(self)

    def and_(self, other) -> LogicalAnd:
        """Return :class:`LogicalAnd` between *self* and *other*.

        .. versionadded:: 2015.2
        """
        return LogicalAnd((self, other))

    def or_(self, other) -> LogicalOr:
        """Return :class:`LogicalOr` between *self* and *other*.

        .. versionadded:: 2015.2
        """
        return LogicalOr((self, other))

    # }}}

    # {{{ comparison constructors

    def eq(self, other) -> Comparison:
        """Return a :class:`Comparison` comparing *self* to *other*.

        .. versionadded:: 2015.2
        """
        return Comparison(self, "==", other)

    def ne(self, other) -> Comparison:
        """Return a :class:`Comparison` comparing *self* to *other*.

        .. versionadded:: 2015.2
        """
        return Comparison(self, "!=", other)

    def le(self, other) -> Comparison:
        """Return a :class:`Comparison` comparing *self* to *other*.

        .. versionadded:: 2015.2
        """
        return Comparison(self, "<=", other)

    def lt(self, other) -> Comparison:
        """Return a :class:`Comparison` comparing *self* to *other*.

        .. versionadded:: 2015.2
        """
        return Comparison(self, "<", other)

    def ge(self, other) -> Comparison:
        """Return a :class:`Comparison` comparing *self* to *other*.

        .. versionadded:: 2015.2
        """
        return Comparison(self, ">=", other)

    def gt(self, other) -> Comparison:
        """Return a :class:`Comparison` comparing *self* to *other*.

        .. versionadded:: 2015.2
        """
        return Comparison(self, ">", other)

    # }}}

    # {{{ prevent less / greater comparisons

    # /!\ Don't be tempted to resolve these to Comparison.

    def __le__(self, other) -> NoReturn:
        raise TypeError("expressions don't have an order")

    def __lt__(self, other) -> NoReturn:
        raise TypeError("expressions don't have an order")

    def __ge__(self, other) -> NoReturn:
        raise TypeError("expressions don't have an order")

    def __gt__(self, other) -> NoReturn:
        raise TypeError("expressions don't have an order")

    # }}}

    def __abs__(self) -> ExpressionNode:
        return Call(Variable("abs"), (self,))

    def __iter__(self):
        # prevent infinite loops (e.g. when inserting into numpy arrays)
        raise TypeError("expression types are not iterable")


# {{{ dataclasses support

# https://stackoverflow.com/a/1176023
_CAMEL_TO_SNAKE_RE = re.compile(
    r"""
        (?<=[a-z])      # preceded by lowercase
        (?=[A-Z])       # followed by uppercase
        |               #   OR
        (?<=[A-Z])      # preceded by lowercase
        (?=[A-Z][a-z])  # followed by uppercase, then lowercase
    """,
    re.X,
)


class _HasMapperMethod(Protocol):
    mapper_method: ClassVar[str]


def _augment_expression_dataclass(
            cls: type[DataclassInstance],
            *,
            generate_eq: bool,
            generate_hash: bool,
        ) -> None:
    attr_tuple = ", ".join(f"self.{fld.name}" for fld in fields(cls))
    if attr_tuple:
        attr_tuple = f"({attr_tuple},)"
    else:
        attr_tuple = "()"

    fld_name_tuple = ", ".join(f"'{fld.name}'" for fld in fields(cls))
    if fld_name_tuple:
        fld_name_tuple = f"({fld_name_tuple},)"
    else:
        fld_name_tuple = "()"

    comparison = " and ".join(
            f"self.{fld.name} == other.{fld.name}"
            for fld in fields(cls))

    if not comparison:
        comparison = "True"

    from pytools.codegen import remove_common_indentation
    augment_code = remove_common_indentation(
        """
        from warnings import warn
        from dataclasses import is_dataclass
        """
        + (f"""


        def {cls.__name__}_eq(self, other):
            if self is other:
                return True
            if self.__class__ is not other.__class__:
                return False
            if {generate_hash}:
                if hash(self) != hash(other):
                    return False
            if self.__class__ is not cls and self.init_arg_names != {fld_name_tuple}:
                warn(f"{{self.__class__}} is derived from {cls}, which is now "
                    f"a dataclass. {{self.__class__}} should be converted to being "
                    "a dataclass as well. Non-dataclass subclasses "
                    "will stop working in 2025.",
                    DeprecationWarning)

                return self.is_equal(other)

            return self.__class__ == other.__class__ and {comparison}

        cls.__eq__ = {cls.__name__}_eq
        """ if generate_eq else "")
        + (f"""


        def {cls.__name__}_hash(self):
            try:
                return self._hash_value
            except AttributeError:
                pass

            if self.__class__ is not cls and self.init_arg_names != {fld_name_tuple}:
                warn(f"{{self.__class__}} is derived from {cls}, which is now "
                    f"a dataclass. {{self.__class__}} should be converted to being "
                    "a dataclass as well. Non-dataclass subclasses "
                    "will stop working in 2025.",
                    DeprecationWarning)

                hash_val = self.get_hash()
            else:
                hash_val = hash({attr_tuple})

            object.__setattr__(self, "_hash_value", hash_val)
            return hash_val

        cls.__hash__ = {cls.__name__}_hash
        """ if generate_hash else "")
        + f"""


        def {cls.__name__}_init_arg_names(self):
            depr_key = (type(self), "init_arg_names")
            if depr_key not in self._deprecation_warnings_issued:
                warn("Attribute 'init_arg_names' of {cls} is deprecated and will "
                        "not have a default implementation starting from 2025. "
                        "Use 'dataclasses.fields' instead.",
                        DeprecationWarning, stacklevel=2)

                self._deprecation_warnings_issued.add(depr_key)

            return {fld_name_tuple}

        cls.init_arg_names = property({cls.__name__}_init_arg_names)


        def {cls.__name__}_getinitargs(self):
            depr_key = (type(self), "__getinitargs__")
            if depr_key not in self._deprecation_warnings_issued:
                warn("Method '__getinitargs__' of {cls} is deprecated and will "
                        "not have a default implementation starting from 2025. "
                        "Use 'dataclasses.fields' instead.",
                        DeprecationWarning, stacklevel=2)

                self._deprecation_warnings_issued.add(depr_key)

            return {attr_tuple}

        cls.__getinitargs__ = {cls.__name__}_getinitargs


        def {cls.__name__}_getstate(self):
            # We might get called on a non-dataclass subclass.
            if "_is_expr_dataclass" not in self.__class__.__dict__:
                from pymbolic.primitives import Expression
                return Expression.__getstate__(self)

            return {attr_tuple}

        cls.__getstate__ = {cls.__name__}_getstate


        def {cls.__name__}_setstate(self, state):
            # We might get called on a non-dataclass subclass.
            if "_is_expr_dataclass" not in self.__class__.__dict__:
                from pymbolic.primitives import Expression
                return Expression.__setstate__(self, state)

            for name, value in zip({fld_name_tuple}, state):
                object.__setattr__(self, name, value)

        cls.__setstate__ = {cls.__name__}_setstate
        """)

    exec_dict = {"cls": cls, "_MODULE_SOURCE_CODE": augment_code}
    exec(compile(augment_code,
                 f"<dataclass augmentation code for {cls}>", "exec"),
         exec_dict)

    # set a marker to detect classes whose subclasses may not be expr_dataclasses
    # type ignore because we don't want to announce the existence of this to the world.
    cls._is_expr_dataclass = True  # type: ignore[attr-defined]

    # {{{ assign mapper_method

    mm_cls = cast(type[_HasMapperMethod], cls)

    snake_clsname = _CAMEL_TO_SNAKE_RE.sub("_", mm_cls.__name__).lower()
    default_mapper_method_name = f"map_{snake_clsname}"

    # This covers two cases: the class does not have the attribute in the first
    # place, or it inherits a value but does not set it itself.
    sets_mapper_method = "mapper_method" in mm_cls.__dict__

    if sets_mapper_method:
        if default_mapper_method_name == mm_cls.mapper_method:
            warn(f"Explicit mapper_method on {mm_cls} not needed, default matches "
                 "explicit assignment. Just delete the explicit assignment.",
                 stacklevel=3)

    if not sets_mapper_method:
        mm_cls.mapper_method = intern(default_mapper_method_name)

    # }}}


_T = TypeVar("_T")


@dataclass_transform(frozen_default=True)
def expr_dataclass(
            init: bool = True,
            eq: bool = True,
            hash: bool = True,
        ) -> Callable[[type[_T]], type[_T]]:
    r"""A class decorator that makes the class a :func:`~dataclasses.dataclass`
    while also adding functionality needed for :class:`ExpressionNode`.
    Specifically, it adds cached hashing, equality comparisons
    with ``self is other`` shortcuts as well as some methods/attributes
    for backward compatibility (e.g. ``__getinitargs__``, ``init_arg_names``).

    It also adds a :attr:`ExpressionNode.mapper_method` based on the class name
    if not already present. If :attr:`~ExpressionNode.mapper_method` is inherited,
    it will be viewed as unset and replaced.

    Note that the class to which this decorator is applied need not be
    a subclass of :class:`ExpressionNode`.

    .. versionadded:: 2024.1
    """
    def map_cls(cls: type[_T]) -> type[_T]:
        # Frozen dataclasses (empirically) have a ~20% speed penalty in pymbolic,
        # and their frozen-ness is arguably a debug feature.

        # We provide __eq__/__hash__ below, don't redundantly generate it.
        dc_cls = dataclass(init=init, eq=False, frozen=__debug__, repr=False)(cls)

        # FIXME: I'm not sure how to tell mypy that dc_cls is type[DataclassInstance]
        # It should just understand that?
        _augment_expression_dataclass(
                  dc_cls,  # type: ignore[arg-type]
                  generate_eq=eq and "__eq__" not in cls.__dict__,
                  generate_hash=hash and "__hash__" not in cls.__dict__,
                  )
        return dc_cls

    return map_cls

# }}}


@expr_dataclass()
class AlgebraicLeaf(ExpressionNode):
    """An expression that serves as a leaf for arithmetic evaluation.
    This may end up having child nodes still, but they're not reached by
    ways of arithmetic."""
    pass


@expr_dataclass()
class Leaf(AlgebraicLeaf):
    """An expression that is irreducible, i.e. has no Expression-type parts
    whatsoever."""
    pass


@expr_dataclass()
class Variable(Leaf):
    """
    .. autoattribute:: name
    """
    name: str


@expr_dataclass()
class Wildcard(Leaf):
    """A general wildcard that can be used to substitute expressions."""


@expr_dataclass()
class DotWildcard(Leaf):
    """A wildcard that can be substituted for a single expression."""
    name: str


@expr_dataclass()
class StarWildcard(Leaf):
    """A wildcard that can be substituted by a sequence of expressions of
    non-negative length.
    """
    name: str


@expr_dataclass()
class FunctionSymbol(AlgebraicLeaf):
    """Represents the name of a function.

    May optionally have an `arg_count` attribute, which will
    allow `Call` to check the number of arguments.
    """


# {{{ structural primitives

@expr_dataclass()
class Call(AlgebraicLeaf):
    """A function invocation.

    .. autoattribute:: function
    .. autoattribute:: parameters
    """
    function: _Expression
    """Evaluates to the function to be called."""

    parameters: tuple[_Expression, ...]
    """
    A :class:`tuple` of positional parameters.
    """


@expr_dataclass()
class CallWithKwargs(AlgebraicLeaf):
    """A function invocation with keyword arguments.

    .. autoattribute:: function
    .. autoattribute:: parameters
    .. autoattribute:: kw_parameters
    """

    function: _Expression
    """Evaluates to the function to be called."""

    parameters: tuple[_Expression, ...]
    """A :class:`tuple` of positional parameters.
    """

    kw_parameters: Mapping[str, _Expression]
    """A dictionary mapping names to arguments.
    """

    def __post_init__(self):
        try:
            hash(self.kw_parameters)
        except Exception:
            warn("CallWithKwargs created with non-hashable kw_parameters. "
                 "This is deprecated and will stop working in 2025. "
                 "If you need an immutable mapping, "
                 "try the :mod:`immutabledict` package.",
                 DeprecationWarning, stacklevel=3
             )
            object.__setattr__(self, "kw_parameters", immutabledict(self.kw_parameters))


@expr_dataclass()
class Subscript(AlgebraicLeaf):
    """An array subscript."""

    aggregate: _Expression
    index: _Expression

    @property
    def index_tuple(self) -> tuple[_Expression, ...]:
        """
        Return :attr:`index` wrapped in a single-element tuple, if it is not already
        a tuple.
        """

        if isinstance(self.index, tuple):
            return self.index
        else:
            return (self.index,)


@expr_dataclass()
class Lookup(AlgebraicLeaf):
    """Access to an attribute of an *aggregate*, such as an attribute of a class."""

    aggregate: _Expression
    name: str

# }}}


# {{{ arithmetic primitives

@expr_dataclass()
class Sum(ExpressionNode):
    """
    .. autoattribute:: children

    .. automethod:: __add__
    .. automethod:: __radd__
    .. automethod:: __sub__
    .. automethod:: __bool__
    """

    children: tuple[_Expression, ...]

    def __add__(self, other):
        if not is_valid_operand(other):
            return NotImplemented

        if isinstance(other, Sum):
            return Sum(self.children + other.children)
        if not other:
            return self
        return Sum((*self.children, other))

    def __radd__(self, other):
        if not is_constant(other):
            return NotImplemented

        if isinstance(other, Sum):
            return Sum(other.children + self.children)
        if not other:
            return self
        return Sum((other, *self.children))

    def __sub__(self, other):
        if not is_valid_operand(other):
            return NotImplemented

        if not other:
            return self
        return Sum((*self.children, -other))

    def __bool__(self):
        if len(self.children) == 0:
            return True
        elif len(self.children) == 1:
            return bool(self.children[0])
        else:
            # FIXME: Right semantics?
            return True

    __nonzero__ = __bool__


@expr_dataclass()
class Product(ExpressionNode):
    """
    .. autoattribute:: children

    .. automethod:: __mul__
    .. automethod:: __rmul__
    .. automethod:: __bool__
    """

    children: tuple[_Expression, ...]

    def __mul__(self, other):
        if not is_valid_operand(other):
            return NotImplemented
        if isinstance(other, Product):
            return Product(self.children + other.children)
        if is_zero(other):
            return 0
        if is_zero(other-1):
            return self
        return Product((*self.children, other))

    def __rmul__(self, other):
        if not is_constant(other):
            return NotImplemented
        if isinstance(other, Product):
            return Product(other.children + self.children)
        if is_zero(other):
            return 0
        if is_zero(other-1):
            return self
        return Product((other, *self.children))

    def __bool__(self):
        for i in self.children:
            if is_zero(i):
                return False
        return True

    __nonzero__ = __bool__


@expr_dataclass()
class Min(ExpressionNode):
    """
    .. autoattribute:: children
    """
    children: tuple[_Expression, ...]


@expr_dataclass()
class Max(ExpressionNode):
    """
    .. autoattribute:: children
    """
    children: tuple[_Expression, ...]


@expr_dataclass()
class QuotientBase(ExpressionNode):
    numerator: ArithmeticExpression
    denominator: ArithmeticExpression

    @property
    def num(self):
        return self.numerator

    @property
    def den(self):
        return self.denominator

    def __bool__(self):
        return bool(self.numerator)

    __nonzero__ = __bool__


@expr_dataclass()
class Quotient(QuotientBase):
    """Bases: :class:`~pymbolic.ExpressionNode`

    .. autoattribute:: numerator
    .. autoattribute:: denominator
    """


@expr_dataclass()
class FloorDiv(QuotientBase):
    """Bases: :class:`~pymbolic.ExpressionNode`

    .. autoattribute:: numerator
    .. autoattribute:: denominator
    """


@expr_dataclass()
class Remainder(QuotientBase):
    """Bases: :class:`~pymbolic.ExpressionNode`

    .. autoattribute:: numerator
    .. autoattribute:: denominator
    """


@expr_dataclass()
class Power(ExpressionNode):
    """
    .. autoattribute:: base
    .. autoattribute:: exponent
    """

    base: ArithmeticExpression
    exponent: ArithmeticExpression

# }}}


# {{{ shift operators

@expr_dataclass()
class _ShiftOperator(ExpressionNode):
    shiftee: _Expression
    shift: _Expression


@expr_dataclass()
class LeftShift(_ShiftOperator):
    """Bases: :class:`~pymbolic.ExpressionNode`.

    .. autoattribute:: shiftee
    .. autoattribute:: shift
    """


@expr_dataclass()
class RightShift(_ShiftOperator):
    """Bases: :class:`~pymbolic.ExpressionNode`.

    .. autoattribute:: shiftee
    .. autoattribute:: shift
    """

# }}}


# {{{ bitwise operators

@expr_dataclass()
class BitwiseNot(ExpressionNode):
    """
    .. autoattribute:: child
    """

    child: _Expression


@expr_dataclass()
class BitwiseOr(ExpressionNode):
    """
    .. autoattribute:: children
    """

    children: tuple[_Expression, ...]


@expr_dataclass()
class BitwiseXor(ExpressionNode):
    """
    .. autoattribute:: children
    """
    children: tuple[_Expression, ...]


@expr_dataclass()
class BitwiseAnd(ExpressionNode):
    """
    .. autoattribute:: children
    """
    children: tuple[_Expression, ...]

# }}}


# {{{ comparisons, logic, conditionals

@expr_dataclass()
class Comparison(ExpressionNode):
    """
    .. autoattribute:: left
    .. autoattribute:: operator
    .. autoattribute:: right

    .. note::

        Unlike other expressions, comparisons are not implicitly constructed by
        comparing :class:`ExpressionNode` objects.
        See :meth:`pymbolic.ExpressionNode.eq` and related.

    .. autoattribute:: operator_to_name
    .. autoattribute:: name_to_operator
    """

    left: _Expression

    operator: str
    """One of ``[">", ">=", "==", "!=", "<", "<="]``."""

    right: _Expression

    operator_to_name: ClassVar[dict[str, str]] = {
            "==": "eq",
            "!=": "ne",
            ">=": "ge",
            ">": "gt",
            "<=": "le",
            "<": "lt",
            }
    name_to_operator: ClassVar[dict[str, str]] = {
        name: op for op, name in operator_to_name.items()
    }

    def __post_init__(self):
        # FIXME Yuck, gross
        if self.operator not in self.operator_to_name:
            if self.operator in self.name_to_operator:
                warn("Passing operators by name is deprecated and will stop working "
                     "in 2025. "
                     "Use the name_to_operator class attribute to translate in "
                     "calling code instead.",
                     DeprecationWarning, stacklevel=3)

                object.__setattr__(
                        self, "operator", self.name_to_operator[self.operator])
            else:
                raise RuntimeError(f"invalid operator: '{self.operator}'")


@expr_dataclass()
class LogicalNot(ExpressionNode):
    """
    .. autoattribute:: child
    """

    child: _Expression


@expr_dataclass()
class LogicalOr(ExpressionNode):
    """
    .. autoattribute:: children
    """

    children: tuple[_Expression, ...]


@expr_dataclass()
class LogicalAnd(ExpressionNode):
    """
    .. autoattribute:: children
    """
    children: tuple[_Expression, ...]


@expr_dataclass()
class If(ExpressionNode):
    """
    .. autoattribute:: condition
    .. autoattribute:: then
    .. autoattribute:: else_
    """

    condition: _Expression
    then: _Expression
    else_: _Expression

# }}}


# {{{ misc stuff

class cse_scope:  # noqa
    """Determines the lifetime for the saved value of a :class:`CommonSubexpression`.

    .. attribute:: EVALUATION
        :type: str

        The evaluated result lives for the duration of the evaluation of the
        current expression and is discarded thereafter.

    .. attribute:: EXPRESSION
        :type: str

        The evaluated result lives for the lifetime of the current expression
        (across multiple evaluations with multiple parameters) and is discarded
        when the expression is.

    .. attribute:: GLOBAL
        :type: str

        The evaluated result lives until the execution context dies.
    """

    EVALUATION = "pymbolic_eval"
    EXPRESSION = "pymbolic_expr"
    GLOBAL = "pymbolic_global"


@expr_dataclass()
class CommonSubexpression(ExpressionNode):
    """A helper for code generation and caching. Denotes a subexpression that
    should only be evaluated once. If, in code generation, it is assigned to
    a variable, a name starting with :attr:`prefix` should be used.

    .. autoattribute:: child
    .. autoattribute:: prefix
    .. autoattribute:: scope


    See :class:`pymbolic.mapper.c_code.CCodeMapper` for an example.
    """

    child: _Expression
    prefix: str | None = None
    scope: str = cse_scope.EVALUATION
    """
    One of the values in :class:`cse_scope`. See there for meaning.
    """

    def __post_init__(self):
        if self.scope is None:
            warn("CommonSubexpression.scope set to None. "
                 "This is deprecated and will stop working in 2024. "
                "Use cse_scope.EVALUATION explicitly instead.",
                 DeprecationWarning, stacklevel=3)
            object.__setattr__(self, "scope", cse_scope.EVALUATION)

    def get_extra_properties(self):
        """Return a dictionary of extra kwargs to be passed to the
        constructor from the identity mapper.

        This allows derived classes to exist without having to
        extend every mapper that processes them.
        """

        return {}


@expr_dataclass()
class Substitution(ExpressionNode):
    """Work-alike of :class:`~sympy.core.function.Subs`.

    .. autoattribute:: child
    .. autoattribute:: variables
    .. autoattribute:: values
    """

    child: _Expression
    variables: tuple[str, ...]
    values: tuple[_Expression, ...]


@expr_dataclass()
class Derivative(ExpressionNode):
    """Work-alike of sympy's :class:`~sympy.core.function.Derivative`.

    .. autoattribute:: child
    .. autoattribute:: variables
    """

    child: _Expression
    variables: tuple[str, ...]


SliceChildrenT: TypeAlias = (tuple[()]
        | tuple[_Expression | None]
        | tuple[_Expression | None, _Expression | None]
        | tuple[_Expression | None, _Expression | None, _Expression
            | None])


@expr_dataclass()
class Slice(ExpressionNode):
    """A slice expression as in a[1:7].

    .. autoattribute:: children

    .. autoproperty:: start
    .. autoproperty:: stop
    .. autoproperty:: step
    """

    children: SliceChildrenT

    def __bool__(self):
        return True

    __nonzero__ = __bool__

    @property
    def start(self):
        if len(self.children) > 0:
            return self.children[0]
        else:
            return None

    @property
    def stop(self):
        if len(self.children) == 1:
            return self.children[0]
        elif len(self.children) > 1:
            return self.children[1]
        else:
            return None

    @property
    def step(self):
        if len(self.children) == 3:
            return self.children[2]
        else:
            return None


@expr_dataclass()
class NaN(AlgebraicLeaf):
    """
    An expression node representing not-a-number as a floating point number.
    Unlike, :data:`math.nan`, all instances of :class:`NaN` compare equal, as
    one might reasonably expect for program representation. (If this weren't
    so, programs containing NaNs would effectively be unhashable, because they
    don't compare equal to themselves.)

    Note that, in Python, this equality comparison is made *even* more
    complex by `this issue <https://bugs.python.org/issue21873>`__, due
    to which ``np.nan == np.nan`` is *False*, but ``(np.nan,) == (np.nan,)``
    is True.

    .. autoattribute:: data_type

        The data type used for the actual realization of the constant. Defaults
        to *None*. If given, This must be a callable to which a NaN
        :class:`float` can be passed to obtain a NaN of the yield the desired
        type.  It must also be suitable for use as the second argument of
        :func:`isinstance`.
    """
    data_type: Callable[[float], Any] | None = None

    mapper_method = intern("map_nan")

# }}}


# {{{ intelligent factory functions

def make_variable(var_or_string: Variable | str) -> Variable:
    if isinstance(var_or_string, str):
        return Variable(intern(var_or_string))
    else:
        return var_or_string


def subscript(expression, index):
    return Subscript(expression, index)


def flattened_sum(terms: Iterable[ArithmeticExpression]) -> ArithmeticExpression:
    r"""Recursively flattens all the top level :class:`Sum`\ s in *terms*.

    :arg terms: an :class:`~collections.abc.Iterable` of expressions.
    :returns: a :class:`Sum` expression or, if there is only one term in
        the sum, the respective term.
    """
    queue = list(terms)
    done = []

    while queue:
        item = queue.pop(0)

        if is_zero(item):
            continue

        if isinstance(item, Sum):
            ch = cast(tuple[ArithmeticExpression], item.children)
            queue.extend(ch)
        else:
            done.append(item)

    if len(done) == 0:
        return 0
    elif len(done) == 1:
        return done[0]
    else:
        return Sum(tuple(done))


def linear_combination(coefficients, expressions):
    return sum(coefficient * expression
                 for coefficient, expression
                 in zip(coefficients, expressions, strict=True)
                 if coefficient and expression)


def flattened_product(terms: Iterable[ArithmeticExpression]) -> ArithmeticExpression:
    r"""Recursively flattens all the top level :class:`Product`\ s in *terms*.

    This operation does not change the order of the terms in the products, so
    it does not require the product to be commutative.

    :arg terms: an :class:`~collections.abc.Iterable` of expressions.
    :returns: a :class:`Product` expression or, if there is only one term in
        the product, the respective term.
    """
    queue = list(terms)
    done = []

    while queue:
        item = queue.pop(0)

        if is_zero(item):
            return 0
        if is_zero(item - 1):
            continue

        if isinstance(item, Product):
            ch = cast(tuple[ArithmeticExpression], item.children)
            queue.extend(ch)
        else:
            done.append(item)

    if len(done) == 0:
        return 1
    elif len(done) == 1:
        return done[0]
    else:
        return Product(tuple(done))


def quotient(numerator, denominator):
    if not (denominator-1):
        return numerator

    import pymbolic.rational as rat
    if isinstance(numerator, rat.Rational) and \
            isinstance(denominator, rat.Rational):
        return numerator * denominator.reciprocal()

    try:
        c_traits = traits.common_traits(numerator, denominator)
        if isinstance(c_traits, traits.EuclideanRingTraits):
            return rat.Rational(numerator, denominator)
    except traits.NoCommonTraitsError:
        pass
    except traits.NoTraitsError:
        pass

    return Quotient(numerator, denominator)

# }}}


# {{{ tool functions

global VALID_CONSTANT_CLASSES
global VALID_OPERANDS
VALID_CONSTANT_CLASSES: tuple[type, ...] = (int, float, complex)
_BOOL_CLASSES: tuple[type, ...] = (bool,)
VALID_OPERANDS = (ExpressionNode,)

try:
    import numpy
    VALID_CONSTANT_CLASSES += (numpy.number, numpy.bool_)
    _BOOL_CLASSES += (numpy.bool_, )
except ImportError:
    pass


def is_constant(value: object) -> TypeIs[Scalar]:
    return isinstance(value, VALID_CONSTANT_CLASSES)


def is_number(value: object) -> TypeIs[Number]:
    return (not isinstance(value, _BOOL_CLASSES)
        and isinstance(value, VALID_CONSTANT_CLASSES))


def is_valid_operand(value: object) -> TypeIs[_Expression]:
    return isinstance(value, VALID_OPERANDS) or is_constant(value)


def is_arithmetic_expression(value: object) -> TypeIs[ArithmeticExpression]:
    return not isinstance(value, _BOOL_CLASSES) and is_valid_operand(value)


def register_constant_class(class_):
    global VALID_CONSTANT_CLASSES

    VALID_CONSTANT_CLASSES += (class_,)


def unregister_constant_class(class_):
    global VALID_CONSTANT_CLASSES

    tmp = list(VALID_CONSTANT_CLASSES)
    tmp.remove(class_)
    VALID_CONSTANT_CLASSES = tuple(tmp)


def is_nonzero(value: object) -> bool:
    if value is None:
        raise ValueError("is_nonzero is undefined for None")

    try:
        return bool(value)
    except ValueError:
        return True


def is_zero(value: object) -> bool:
    return not is_nonzero(value)


def wrap_in_cse(expr: _Expression,
                prefix: str | None = None,
                scope: str | None = None) -> _Expression:
    warn("'wrap_in_cse' is deprecated and will be removed in 2025. Use "
         "'make_common_subexpression' with the `wrap_vars=False` flag instead.",
         DeprecationWarning, stacklevel=2)

    return make_common_subexpression(expr, prefix, scope, wrap_vars=False)


def make_common_subexpression(expr: _Expression,
                              prefix: str | None = None,
                              scope: str | None = None,
                              *,
                              wrap_vars: bool = True) -> _Expression:
    """Wrap *expr* in a :class:`CommonSubexpression` with *prefix*.

    If *expr* is a :mod:`numpy` object array, each individual entry is instead
    wrapped. If *expr* is a :class:`pymbolic.geometric_algebra.MultiVector`, each
    coefficient is individually wrapped. In general, the function tries to avoid
    re-wrapping existing :class:`CommonSubexpression` if the same scope is given.

    See :class:`CommonSubexpression` for the meaning of *prefix* and *scope*. The
    scope defaults to :attr:`cse_scope.EVALUATION`.

    :arg wrap_vars: If *True*, this also wraps a :class:`~pymbolic.primitives.Variable`
        and its subscripts. Otherwise, these expressions are returned as is.
    """

    if is_constant(expr):
        return expr

    if not wrap_vars and isinstance(expr, Variable | Subscript):
        return expr

    # handle CSE re-wrapping
    if scope is None:
        scope = cse_scope.EVALUATION

    if isinstance(expr, CommonSubexpression):
        if scope == cse_scope.EVALUATION or expr.scope == scope:
            return expr

    # handle MultiVector
    from pymbolic.geometric_algebra import MultiVector

    if isinstance(expr, MultiVector):
        new_data = {}
        for bits, coeff in expr.data.items():
            if prefix is not None:
                blade_str = expr.space.blade_bits_to_str(bits, "")
                component_prefix = f"{prefix}_{blade_str}"
            else:
                component_prefix = None

            new_data[bits] = make_common_subexpression(
                    coeff, component_prefix, scope, wrap_vars=wrap_vars)

        return MultiVector(new_data, expr.space)

    # handle numpy object arrays
    try:
        import numpy as np

        if isinstance(expr, np.ndarray) and expr.dtype.char == "O":
            is_obj_array = True
            logical_shape = expr.shape
        else:
            is_obj_array = False
            logical_shape = ()
    except ImportError:
        is_obj_array = False
        logical_shape = ()

    if is_obj_array and logical_shape != ():
        assert isinstance(expr, np.ndarray)

        result = np.zeros(logical_shape, dtype=object)
        for i in np.ndindex(logical_shape):
            if prefix is not None:
                bits = "_".join(str(i_i) for i_i in i)
                component_prefix = f"{prefix}_{bits}"
            else:
                component_prefix = None

            result[i] = make_common_subexpression(
                    expr[i], component_prefix, scope, wrap_vars=wrap_vars)

        return result

    # everything else gets re-wrapped
    return CommonSubexpression(expr, prefix, scope)


def make_sym_vector(name, components, var_factory=Variable):
    """Return an object array of *components* subscripted
    :class:`Variable` (or subclass) instances.

    :arg components: Either a list of indices, or an integer representing the
        number of indices.
    :arg var_factory: The :class:`Variable` subclass to
        use for instantiating the scalar variables.

    For example, this creates a vector with three components::

        >>> make_sym_vector("vec", 3)
        array([Subscript(Variable('vec'), 0), Subscript(Variable('vec'), 1),
               Subscript(Variable('vec'), 2)], dtype=object)

    """
    from numbers import Integral
    if isinstance(components, Integral):
        components = list(range(components))

    from pytools.obj_array import flat_obj_array
    vfld = var_factory(name)
    return flat_obj_array(*[vfld[i] for i in components])


def make_sym_array(name, shape, var_factory=Variable):
    vfld = var_factory(name)
    if shape == ():
        return vfld

    import numpy as np
    result = np.zeros(shape, dtype=object)
    for i in np.ndindex(shape):
        result[i] = vfld[i]

    return result


def variables(s):
    """Return a list of variables for each (space-delimited) identifier
    in *s*.
    """
    return [Variable(s_i) for s_i in s.split() if s_i]

# }}}


__getattr__ = partial(module_getattr_for_deprecations, __name__, {
        "Expression": ("ExpressionNode", ExpressionNode, 2026),
        })


# vim: foldmethod=marker



from __future__ import annotations


__copyright__ = "Copyright (C) 2014 Andreas Kloeckner"

__license__ = """
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""

# This is experimental, undocumented, and could go away any second.
# Consider yourself warned.
from collections.abc import Set
from typing import ClassVar

import pymbolic.geometric_algebra.primitives as prim
from pymbolic.geometric_algebra import MultiVector
from pymbolic.mapper import (
    CachedMapper,
    CollectedT,
    Collector as CollectorBase,
    CombineMapper as CombineMapperBase,
    IdentityMapper as IdentityMapperBase,
    P,
    ResultT,
    WalkMapper as WalkMapperBase,
)
from pymbolic.mapper.constant_folder import (
    ConstantFoldingMapper as ConstantFoldingMapperBase,
)
from pymbolic.mapper.evaluator import EvaluationMapper as EvaluationMapperBase
from pymbolic.mapper.graphviz import GraphvizMapper as GraphvizMapperBase
from pymbolic.mapper.stringifier import (
    PREC_NONE,
    StringifyMapper as StringifyMapperBase,
)
from pymbolic.primitives import ExpressionNode


class IdentityMapper(IdentityMapperBase[P]):
    def map_nabla(
                self, expr: prim.Nabla, *args: P.args, **kwargs: P.kwargs
            ) -> ExpressionNode:
        return expr

    def map_nabla_component(self,
                expr: prim.NablaComponent, *args: P.args, **kwargs: P.kwargs
            ) -> ExpressionNode:
        return expr

    def map_derivative_source(self,
                expr: prim.DerivativeSource, *args: P.args, **kwargs: P.kwargs
            ) -> ExpressionNode:
        operand = self.rec(expr.operand, *args, **kwargs)
        if operand is expr.operand:
            return expr

        return type(expr)(operand, expr.nabla_id)


class CombineMapper(CombineMapperBase[ResultT, P]):
    def map_derivative_source(
                self, expr: prim.DerivativeSource, *args: P.args, **kwargs: P.kwargs
            ) -> ResultT:
        return self.rec(expr.operand, *args, **kwargs)


class Collector(CollectorBase[CollectedT, P]):
    def map_nabla(self,
                expr: prim.Nabla, *args: P.args, **kwargs: P.kwargs
            ) -> Set[CollectedT]:
        return set()

    def map_nabla_component(self,
                expr: prim.NablaComponent, *args: P.args, **kwargs: P.kwargs
            ) -> Set[CollectedT]:
        return set()


class WalkMapper(WalkMapperBase[P]):
    def map_nabla(self, expr: prim.Nabla, *args: P.args, **kwargs: P.kwargs) -> None:
        self.visit(expr, *args, **kwargs)
        self.post_visit(expr, *args, **kwargs)

    def map_nabla_component(
                self, expr: prim.NablaComponent, *args: P.args, **kwargs: P.kwargs
            ) -> None:
        self.visit(expr, *args, **kwargs)
        self.post_visit(expr, *args, **kwargs)

    def map_derivative_source(
                self, expr, *args: P.args, **kwargs: P.kwargs
            ) -> None:
        if not self.visit(expr, *args, **kwargs):
            return

        self.rec(expr.operand, *args, **kwargs)
        self.post_visit(expr, *args, **kwargs)


class EvaluationMapper(EvaluationMapperBase):
    def map_nabla_component(self, expr):
        return expr

    map_nabla = map_nabla_component

    def map_derivative_source(self, expr):
        operand = self.rec(expr.operand)
        if operand is expr.operand:
            return expr

        return type(expr)(operand, expr.nabla_id)


class StringifyMapper(StringifyMapperBase[[]]):
    AXES: ClassVar[dict[int, str]] = {0: "x", 1: "y", 2: "z"}

    def map_nabla(self, expr, enclosing_prec):
        return f"∇[{expr.nabla_id}]"

    def map_nabla_component(self, expr, enclosing_prec):
        return "∇{}[{}]".format(
                self.AXES.get(expr.ambient_axis, expr.ambient_axis),
                expr.nabla_id)

    def map_derivative_source(self, expr, enclosing_prec):
        return r"D[{}]({})".format(expr.nabla_id, self.rec(expr.operand, PREC_NONE))


class GraphvizMapper(GraphvizMapperBase):
    def map_derivative_source(self, expr):
        self.lines.append(
                '{} [label="D[{}]",shape=ellipse];'.format(
                    self.get_id(expr), expr.nabla_id))
        if not self.visit(expr, node_printed=True):
            return

        self.rec(expr.operand)
        self.post_visit(expr)


# {{{ dimensionalizer

class Dimensionalizer(EvaluationMapper):
    """
    .. attribute:: ambient_dim

        Dimension of ambient space. Must be provided by subclass.
    """

    @property
    def ambient_dim(self):
        raise NotImplementedError

    def map_multivector_variable(self, expr):
        from pymbolic.primitives import make_sym_vector
        return MultiVector(
                make_sym_vector(expr.name, self.ambient_dim,
                    var_factory=type(expr)))

    def map_nabla(self, expr):
        from pytools.obj_array import make_obj_array
        return MultiVector(make_obj_array(
            [prim.NablaComponent(axis, expr.nabla_id)
                for axis in range(self.ambient_dim)]))

    def map_derivative_source(self, expr):
        rec_op = self.rec(expr.operand)

        if isinstance(rec_op, MultiVector):
            from pymbolic.geometric_algebra.primitives import DerivativeSource
            return rec_op.map(
                    lambda coeff: DerivativeSource(coeff, expr.nabla_id))
        else:
            return super().map_derivative_source(expr)

# }}}


# {{{ derivative binder

class DerivativeSourceAndNablaComponentCollector(CachedMapper, Collector):
    def __init__(self) -> None:
        Collector.__init__(self)
        CachedMapper.__init__(self)

    def map_nabla(self, expr):
        raise RuntimeError("DerivativeOccurrenceMapper must be invoked after "
                "Dimensionalizer--Nabla found, not allowed")

    def map_nabla_component(self, expr):
        return {expr}

    def map_derivative_source(self, expr):
        return {expr} | self.rec(expr.operand)


class NablaComponentToUnitVector(EvaluationMapper):
    def __init__(self, nabla_id, ambient_axis):
        self.nabla_id = nabla_id
        self.ambient_axis = ambient_axis

    def map_variable(self, expr):
        return expr

    def map_nabla_component(self, expr):
        if expr.nabla_id == self.nabla_id:
            if expr.ambient_axis == self.ambient_axis:
                return 1
            else:
                return 0
        else:
            return EvaluationMapper.map_nabla_component(self, expr)


class DerivativeSourceFinder(EvaluationMapper):
    """Recurses down until it finds the
    :class:`pymbolic.geometric_algebra.primitives.DerivativeSource`
    with the right *nabla_id*, then calls :method:`DerivativeBinder.take_derivative`
    on the source's argument.
    """

    def __init__(self, nabla_id, binder, ambient_axis):
        self.nabla_id = nabla_id
        self.binder = binder
        self.ambient_axis = ambient_axis

    def map_derivative_source(self, expr):
        if expr.nabla_id == self.nabla_id:
            return self.binder.take_derivative(self.ambient_axis, expr.operand)
        else:
            return EvaluationMapper.map_derivative_source(self, expr)


class DerivativeBinder(IdentityMapper):
    derivative_source_and_nabla_component_collector = \
            DerivativeSourceAndNablaComponentCollector
    nabla_component_to_unit_vector = NablaComponentToUnitVector
    derivative_source_finder = DerivativeSourceFinder

    def __init__(self, restrict_to_id=None):
        self.derivative_collector = \
                self.derivative_source_and_nabla_component_collector()
        self.restrict_to_id = restrict_to_id

    def take_derivative(self, ambient_axis, expr):
        raise NotImplementedError

    def map_product(self, expr):
        # {{{ gather NablaComponents and DerivativeSources

        d_source_nabla_ids_per_child = []

        # id to set((child index, axis), ...)
        nabla_finder = {}
        has_d_source_nablas = False

        for child_idx, child in enumerate(expr.children):
            d_or_ns = self.derivative_collector(child)
            if not d_or_ns:
                d_source_nabla_ids_per_child.append(set())
                continue

            nabla_component_ids = set()
            derivative_source_ids = set()

            nablas = []
            for d_or_n in d_or_ns:
                if isinstance(d_or_n, prim.NablaComponent):
                    nabla_component_ids.add(d_or_n.nabla_id)
                    nablas.append(d_or_n)
                elif isinstance(d_or_n, prim.DerivativeSource):
                    derivative_source_ids.add(d_or_n.nabla_id)
                else:
                    raise RuntimeError("unexpected result from "
                            "DerivativeSourceAndNablaComponentCollector")

            d_source_nabla_ids_per_child.append(derivative_source_ids)
            if derivative_source_ids:
                has_d_source_nablas = True

            for ncomp in nablas:
                nabla_finder.setdefault(
                        ncomp.nabla_id, set()).add((child_idx, ncomp.ambient_axis))

        if nabla_finder and not any(d_source_nabla_ids_per_child):
            raise ValueError(f"no derivative source found to resolve in '{expr}'"
                    " -- did you forget to wrap the term that should have its "
                    "derivative taken in 'Derivative()(term)'?")

        if not has_d_source_nablas:
            rec_children = [self.rec(child) for child in expr.children]
            if all(rec_child is child
                    for rec_child, child in zip(
                            rec_children, expr.children, strict=True)):
                return expr

            return type(expr)(tuple(rec_children))

        # }}}

        # a list of lists, the outer level presenting a sum, the inner a product
        result = [list(expr.children)]

        for child_idx, (d_source_nabla_ids, _child) in enumerate(
                zip(d_source_nabla_ids_per_child, expr.children, strict=True)):
            if not d_source_nabla_ids:
                continue

            if len(d_source_nabla_ids) > 1:
                raise NotImplementedError("more than one DerivativeSource per "
                        "child in a product")

            nabla_id, = d_source_nabla_ids
            try:
                nablas = nabla_finder[nabla_id]
            except KeyError:
                continue

            if self.restrict_to_id is not None and nabla_id != self.restrict_to_id:
                continue

            n_axes = max(axis for _, axis in nablas) + 1

            new_result = []
            for prod_term_list in result:
                for axis in range(n_axes):
                    new_ptl = prod_term_list[:]
                    dsfinder = self.derivative_source_finder(nabla_id, self, axis)

                    new_ptl[child_idx] = dsfinder(new_ptl[child_idx])
                    for nabla_child_index, _ in nablas:
                        new_ptl[nabla_child_index] = \
                                self.nabla_component_to_unit_vector(nabla_id, axis)(
                                        new_ptl[nabla_child_index])

                    new_result.append(new_ptl)

            result = new_result

        from pymbolic.primitives import flattened_product, flattened_sum
        return flattened_sum([
                    flattened_product([
                        self.rec(prod_term) for prod_term in prod_term_list
                        ])
                    for prod_term_list in result
                    ])

    map_bitwise_xor = map_product
    map_bitwise_or = map_product
    map_left_shift = map_product
    map_right_shift = map_product

    def map_derivative_source(self, expr):
        rec_operand = self.rec(expr.operand)

        nablas = []
        for d_or_n in self.derivative_collector(rec_operand):
            if isinstance(d_or_n, prim.NablaComponent):
                nablas.append(d_or_n)
            elif isinstance(d_or_n, prim.DerivativeSource):
                pass
            else:
                raise RuntimeError("unexpected result from "
                        "DerivativeSourceAndNablaComponentCollector")

        n_axes = max(n.ambient_axis for n in nablas) + 1
        assert n_axes

        from pymbolic.primitives import flattened_sum
        return flattened_sum([
                self.take_derivative(
                    axis,
                    self.nabla_component_to_unit_vector(expr.nabla_id, axis)
                    (rec_operand))
                for axis in range(n_axes)
                ])

# }}}


class ConstantFoldingMapper(IdentityMapper, ConstantFoldingMapperBase):
    pass

# vim: foldmethod=marker



from __future__ import annotations

from importlib import metadata


def _parse_version(version: str) -> tuple[tuple[int, ...], str]:
    import re

    m = re.match("^([0-9.]+)([a-z0-9]*?)$", VERSION_TEXT)
    assert m is not None

    return tuple(int(nr) for nr in m.group(1).split(".")), m.group(2)


VERSION_TEXT = metadata.version("pymbolic")
VERSION, VERSION_STATUS = _parse_version(VERSION_TEXT)



"""Instruction types"""
from __future__ import annotations


__copyright__ = "Copyright (C) 2015 Matt Wala, Andreas Kloeckner"

__license__ = """
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""

from sys import intern

from pytools import RecordWithoutPickling

from pymbolic.typing import not_none


# {{{ statemetn classes

class Statement(RecordWithoutPickling):
    """
    .. attribute:: depends_on

        A :class:`frozenset` of instruction ids that are reuqired to be
        executed within this execution context before this instruction can be
        executed.

    .. attribute:: id

        A string, a unique identifier for this instruction.

    .. automethod:: get_written_variables
    .. automethod:: get_read_variables
    """

    def __init__(self, **kwargs):
        id = kwargs.pop("id", None)
        if id is not None:
            id = intern(id)

        depends_on = frozenset(kwargs.pop("depends_on", []))
        super().__init__(
                                       id=id,
                                       depends_on=depends_on,
                                       **kwargs)

    def get_written_variables(self):
        """Returns a :class:`frozenset` of variables being written by this
        instruction.
        """
        return frozenset()

    def get_read_variables(self):
        """Returns a :class:`frozenset` of variables being read by this
        instruction.
        """
        return frozenset()

    def map_expressions(self, mapper, include_lhs=True):
        """Returns a new copy of *self* with all expressions
        replaced by ``mapepr(expr)`` for every
        :class:`pymbolic.primitives.Expression`
        contained in *self*.
        """
        return self

    def get_dependency_mapper(self, include_calls="descend_args"):
        from pymbolic.mapper.dependency import DependencyMapper
        return DependencyMapper(
            include_subscripts=False,
            include_lookups=False,
            include_calls=include_calls)

# }}}


# {{{ statement with condition

class ConditionalStatement(Statement):
    __doc__ = not_none(Statement.__doc__) + """
    .. attribute:: condition

       The instruction condition as a :mod:`pymbolic` expression (`True` if the
       instruction is unconditionally executed)
    """

    def __init__(self, **kwargs):
        condition = kwargs.pop("condition", True)
        super().__init__(
                condition=condition,
                **kwargs)

    def _condition_printing_suffix(self):
        if self.condition is True:
            return ""
        return " if " + str(self.condition)

    def __str__(self):
        return (super().__str__()
                + self._condition_printing_suffix())

    def get_read_variables(self):
        dep_mapper = self.get_dependency_mapper()
        return (
                super().get_read_variables()
                | frozenset(
                    dep.name for dep in dep_mapper(self.condition)))

# }}}


# {{{ assignment

class Assignment(Statement):
    """
    .. attribute:: lhs
    .. attribute:: rhs
    """

    def __init__(self, lhs, rhs, **kwargs):
        super().__init__(
                lhs=lhs,
                rhs=rhs,
                **kwargs)

    def get_written_variables(self):
        from pymbolic.primitives import Subscript, Variable
        if isinstance(self.lhs, Variable):
            return frozenset([self.lhs.name])
        elif isinstance(self.lhs, Subscript):
            assert isinstance(self.lhs.aggregate, Variable)
            return frozenset([self.lhs.aggregate.name])
        else:
            raise TypeError("unexpected type of LHS")

    def get_read_variables(self):
        result = super().get_read_variables()
        get_deps = self.get_dependency_mapper()

        def get_vars(expr):
            return frozenset(dep.name for dep in get_deps(self.rhs))

        result = get_vars(self.rhs) | get_vars(self.lhs)

        return result

    def map_expressions(self, mapper, include_lhs=True):
        return (super()
                .map_expressions(mapper, include_lhs=include_lhs)
                .copy(
                    lhs=mapper(self.lhs) if include_lhs else self.lhs,
                    rhs=mapper(self.rhs)))

    def __str__(self):
        result = "{assignee} <- {expr}".format(
            assignee=str(self.lhs),
            expr=str(self.rhs),)

        return result

# }}}


# {{{ conditional assignment

class ConditionalAssignment(ConditionalStatement, Assignment):
    def map_expressions(self, mapper, include_lhs=True):
        return (super()
                .map_expressions(mapper, include_lhs=include_lhs)
                .copy(condition=mapper(self.condition)))

# }}}


# {{{ nop

class Nop(Statement):
    def __str__(self):
        return "nop"

# }}}


Instruction = Statement
ConditionalInstruction = ConditionalStatement


# vim: foldmethod=marker



from __future__ import annotations


__copyright__ = "Copyright (C) 2017 Matt Wala"

__license__ = """
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""

import pytest

import pymbolic.primitives as prim


x_, y_, i_, j_ = (prim.Variable(s) for s in "x y i j".split())


# {{{ to pymbolic test

def _test_to_pymbolic(mapper, sym, use_symengine):
    x, y = sym.symbols("x,y")

    assert mapper(sym.Rational(3, 4)) == prim.Quotient(3, 4)
    assert mapper(sym.Integer(6)) == 6

    if not use_symengine:
        assert mapper(sym.Subs(x**2, (x,), (y,))) == \
            prim.Substitution(x_**2, ("x",), (y_,))
        deriv = sym.Derivative(x**2, x)
        assert mapper(deriv) == prim.Derivative(x_**2, ("x",))
    else:
        assert mapper(sym.Subs(x**2, (x,), (y,))) == \
            y_**2
        deriv = sym.Derivative(x**2, x)
        assert mapper(deriv) == 2*x_

    # functions
    assert mapper(sym.Function("f")(x)) == prim.Variable("f")(x_)
    assert mapper(sym.exp(x)) == prim.Variable("exp")(x_)

    # indexed accesses
    i, j = sym.symbols("i,j")
    if not use_symengine:
        idx = sym.Indexed(x, i, j)
    else:
        idx = sym.Function("Indexed")(x, i, j)
    assert mapper(idx) == x_[i_, j_]

    # constants
    import math
    # FIXME: Why isn't this exact?
    assert abs(mapper(sym.pi) - math.pi) < 1e-14
    assert abs(mapper(sym.E) - math.e) < 1e-14
    assert mapper(sym.I) == 1j

# }}}


def test_symengine_to_pymbolic():
    sym = pytest.importorskip("symengine")
    from pymbolic.interop.symengine import SymEngineToPymbolicMapper
    mapper = SymEngineToPymbolicMapper()

    _test_to_pymbolic(mapper, sym, True)


def test_sympy_to_pymbolic():
    sym = pytest.importorskip("sympy")
    from pymbolic.interop.sympy import SympyToPymbolicMapper
    mapper = SympyToPymbolicMapper()

    _test_to_pymbolic(mapper, sym, False)


# {{{ from pymbolic test

def _test_from_pymbolic(mapper, sym, use_symengine):
    x, y = sym.symbols("x,y")

    assert mapper(x_ + y_) == x + y
    assert mapper(x_ * y_) == x * y
    assert mapper(x_ ** 2) == x ** 2

    assert mapper(prim.Substitution(x_**2, ("x",), (y_,))) == \
        sym.Subs(x**2, (x,), (y,))
    deriv = sym.Derivative(x**2, x)
    assert mapper(prim.Derivative(x_**2, ("x",))) == deriv
    floordiv = sym.floor(x / y)
    assert mapper(prim.FloorDiv(x_, y_)) == floordiv

    if use_symengine:
        assert mapper(x_[0]) == sym.Function("Indexed")("x", 0)
    else:
        i, j = sym.symbols("i,j")
        assert mapper(x_[i_, j_]) == sym.Indexed(x, i, j)

    assert mapper(prim.Variable("f")(x_)) == sym.Function("f")(x)

# }}}


def test_pymbolic_to_symengine():
    sym = pytest.importorskip("symengine")
    from pymbolic.interop.symengine import PymbolicToSymEngineMapper
    mapper = PymbolicToSymEngineMapper()

    _test_from_pymbolic(mapper, sym, True)


def test_pymbolic_to_sympy():
    sym = pytest.importorskip("sympy")
    from pymbolic.interop.sympy import PymbolicToSympyMapper
    mapper = PymbolicToSympyMapper()

    _test_from_pymbolic(mapper, sym, False)


# {{{ roundtrip tests

def _test_roundtrip(forward, backward, sym, use_symengine):
    exprs = [
        2 + x_,
        2 * x_,
        x_ ** 2,
        x_[0],
        x_[i_, j_],
        prim.Variable("f")(x_),
        prim.If(prim.Comparison(x_, "<=", y_), 1, 0),
    ]

    for expr in exprs:
        assert expr == backward(forward(expr))

# }}}


def test_pymbolic_to_sympy_roundtrip():
    sym = pytest.importorskip("sympy")
    from pymbolic.interop.sympy import PymbolicToSympyMapper, SympyToPymbolicMapper
    forward = PymbolicToSympyMapper()
    backward = SympyToPymbolicMapper()

    _test_roundtrip(forward, backward, sym, False)


def test_pymbolic_to_symengine_roundtrip():
    sym = pytest.importorskip("symengine")
    from pymbolic.interop.symengine import (
        PymbolicToSymEngineMapper,
        SymEngineToPymbolicMapper,
    )
    forward = PymbolicToSymEngineMapper()
    backward = SymEngineToPymbolicMapper()

    _test_roundtrip(forward, backward, sym, True)


if __name__ == "__main__":
    import sys
    if len(sys.argv) > 1:
        exec(sys.argv[1])
    else:
        from pytest import main
        main([__file__])



from __future__ import annotations


__copyright__ = """
Copyright (C) 2013 Andreas Kloeckner
Copyright (C) 2014 Matt Wala
"""

__license__ = """
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""

import logging


logger = logging.getLogger(__name__)


# {{{ graphviz / dot export

def _default_preamble_hook():
    # Sets default attributes for nodes and edges.
    yield 'node [shape="box"];'
    yield 'edge [dir="back"];'


def get_dot_dependency_graph(
        statements,  use_stmt_ids=None,
        preamble_hook=_default_preamble_hook,
        additional_lines_hook=list,
        statement_stringifier=None,

        # deprecated
        use_insn_ids=None,):
    """Return a string in the `dot <https://graphviz.org/>`__ language depicting
    dependencies among kernel statements.

    :arg statements: A sequence of statements, each of which is stringified by
        calling *statement_stringifier*.
    :arg statement_stringifier: The function to use for stringifying the
        statements. The default stringifier uses :class:`str` and escapes all
        double quotes (``"``) in the string representation.
    :arg preamble_hook: A function that returns an iterable of lines
        to add at the beginning of the graph
    :arg additional_lines_hook: A function that returns an iterable
        of lines to add at the end of the graph

    """
    if statement_stringifier is None:
        def statement_stringifier(s):
            return str(s).replace('"', r'\"')

    if use_stmt_ids is not None and use_insn_ids is not None:
        raise TypeError("may not specify both use_stmt_ids and use_insn_ids")

    if use_insn_ids is not None:
        use_stmt_ids = use_insn_ids
        from warnings import warn
        warn("'use_insn_ids' is deprecated. Use 'use_stmt_ids' instead.",
                DeprecationWarning, stacklevel=2)

    def get_node_attrs(stmt):
        if use_stmt_ids:
            stmt_label = stmt.id
            tooltip = statement_stringifier(stmt)
        else:
            stmt_label = statement_stringifier(stmt)
            tooltip = stmt.id

        return f'label="{stmt_label}",shape="box",tooltip="{tooltip}"'

    lines = list(preamble_hook())
    lines.append("rankdir=BT;")
    dep_graph = {}

    # maps (oriented) edge onto annotation string
    annotation_dep_graph = {}

    for stmt in statements:
        lines.append('"{}" [{}];'.format(stmt.id, get_node_attrs(stmt)))
        for dep in stmt.depends_on:
            dep_graph.setdefault(stmt.id, set()).add(dep)

        if 0:
            for dep in stmt.then_depends_on:
                annotation_dep_graph[stmt.id, dep] = "then"
            for dep in stmt.else_depends_on:
                annotation_dep_graph[stmt.id, dep] = "else"

    # {{{ O(n^3) (i.e. slow) transitive reduction

    # first, compute transitive closure by fixed point iteration
    while True:
        changed_something = False

        for stmt_1 in dep_graph:
            for stmt_2 in dep_graph.get(stmt_1, set()).copy():
                for stmt_3 in dep_graph.get(stmt_2, set()).copy():
                    if stmt_3 not in dep_graph.get(stmt_1, set()):
                        changed_something = True
                        dep_graph[stmt_1].add(stmt_3)

        if not changed_something:
            break

    for stmt_1 in dep_graph:
        for stmt_2 in dep_graph.get(stmt_1, set()).copy():
            for stmt_3 in dep_graph.get(stmt_2, set()).copy():
                if stmt_3 in dep_graph.get(stmt_1, set()):
                    dep_graph[stmt_1].remove(stmt_3)

    # }}}

    for stmt_1 in dep_graph:
        for stmt_2 in dep_graph.get(stmt_1, set()):
            lines.append(f"{stmt_1} -> {stmt_2}")

    for (stmt_1, stmt_2), annot in annotation_dep_graph.items():
        lines.append(f'{stmt_2} -> {stmt_1}  [label="{annot}", style="dashed"]')

    lines.extend(additional_lines_hook())

    return "digraph code {\n%s\n}" % ("\n".join(lines))

# }}}


# {{{ graphviz / dot interactive show

def show_dot(dot_code, output_to=None):
    from warnings import warn
    warn("pymbolic.imperative.utils.show_dot is deprecated. "
         "It will stop working in July 2023. "
         "Please use pytools.graphviz.show_dot instead.",
         DeprecationWarning, stacklevel=2)

    from pytools.graphviz import show_dot
    return show_dot(dot_code, output_to)
# }}}

# vim: fdm=marker



"""
.. currentmodule:: pymbolic

Typing helpers
--------------

.. autoclass:: Bool
.. autoclass:: Number
.. autoclass:: Scalar
.. autoclass:: ArithmeticExpression

    A narrower type alias than :class:`Expression` that is returned by
    arithmetic operators, to allow continue doing arithmetic with the result
    of arithmetic.

.. currentmodule:: pymbolic.typing

.. autoclass:: Expression

.. note::

    For backward compatibility, ``pymbolic.Expression``
    will alias :class:`pymbolic.primitives.ExpressionNode` for now. Once its deprecation
    period is up, it will be removed, and then, in the further future,
    ``pymbolic.Expression`` may become this type alias.

.. autoclass:: ArithmeticOrExpressionT

    A type variable that can be either :data:`ArithmeticExpression`
    or :data:`Expression`.
"""

from __future__ import annotations


__copyright__ = "Copyright (C) 2024 University of Illinois Board of Trustees"

__license__ = """
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""

from functools import partial
from typing import TYPE_CHECKING, TypeAlias, TypeVar, Union

from pytools import module_getattr_for_deprecations


# FIXME: This is a lie. Many more constant types (e.g. numpy and such)
# are in practical use and completely fine. We cannot really add in numpy
# as a special case (because pymbolic doesn't have a hard numpy dependency),
# and there isn't a usable numerical tower that we could rely on. As such,
# code abusing what constants are allowable will have to type-ignore those
# statements. Better ideas would be most welcome.
#
# References:
# https://github.com/python/mypy/issues/3186
# https://discuss.python.org/t/numeric-generics-where-do-we-go-from-pep-3141-and-present-day-mypy/17155/14

# FIXME: Maybe we should define scalar types via a protocol?
# Pymbolic doesn't particularly restrict scalars to these, this
# just reflects common usage. See typeshed for possible inspiration:
# https://github.com/python/typeshed/blob/119cd09655dcb4ed7fb2021654ba809b8d88846f/stdlib/numbers.pyi

if TYPE_CHECKING:
    from pymbolic.primitives import ExpressionNode

# Experience with depending packages showed that including Decimal and Fraction
# from the stdlib was more trouble than it's worth because those types don't cleanly
# interoperate with each other.
# (e.g. 'Unsupported operand types for * ("Decimal" and "Fraction")')
# And leaving them out doesn't really make any of this more precise.

_StdlibInexactNumberT = float | complex


if TYPE_CHECKING:
    # Yes, type-checking pymbolic will require numpy. That's OK.
    import numpy as np
    Bool = bool | np.bool_
    Integer: TypeAlias = int | np.integer
    InexactNumber: TypeAlias = _StdlibInexactNumberT | np.inexact
else:
    try:
        import numpy as np
    except ImportError:
        Bool = bool
        Integer: TypeAlias = int
        InexactNumber: TypeAlias = _StdlibInexactNumberT
    else:
        Bool = bool | np.bool_
        Integer: TypeAlias = int | np.integer
        InexactNumber: TypeAlias = _StdlibInexactNumberT | np.inexact


Number: TypeAlias = Integer | InexactNumber
Scalar: TypeAlias = Number | Bool

_ScalarOrExpression = Union[Scalar, "ExpressionNode"]
ArithmeticExpression: TypeAlias = Union[Number, "ExpressionNode"]

Expression: TypeAlias = _ScalarOrExpression | tuple["Expression", ...]

ArithmeticOrExpressionT = TypeVar(
                "ArithmeticOrExpressionT",
                ArithmeticExpression,
                Expression)


T = TypeVar("T")


__getattr__ = partial(module_getattr_for_deprecations, __name__, {
        "ArithmeticExpressionT": ("ArithmeticExpression", ArithmeticExpression, 2026),
        "ExpressionT": ("Expression", Expression, 2026),
        "IntegerT": ("Integer", Integer, 2026),
        "ScalarT": ("Scalar", Scalar, 2026),
        "BoolT": ("Bool", Bool, 2026),
        })


def not_none(x: T | None) -> T:
    assert x is not None
    return x



from __future__ import annotations

from collections.abc import Callable
from typing import Any

from pymbolic.interop.matchpy import PymbolicOp


class Mapper:
    def __init__(self) -> None:
        self.cache: dict[PymbolicOp, Any] = {}

    def rec(self, expr: PymbolicOp) -> Any:
        if expr in self.cache:
            return self.cache[expr]

        method: Callable[[PymbolicOp], Any] = getattr(self, expr._mapper_method)

        return method(expr)

    __call__ = rec



"""
.. autofunction:: integer_power
.. autofunction:: extended_euclidean
.. autofunction:: gcd
.. autofunction:: lcm
.. autofunction:: fft
.. autofunction:: ifft
.. autofunction:: sym_fft
.. autofunction:: reduced_row_echelon_form
.. autofunction:: solve_affine_equations_for
"""

from __future__ import annotations


__copyright__ = "Copyright (C) 2009-2013 Andreas Kloeckner"

__license__ = """
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""

import operator
import sys
from typing import TYPE_CHECKING, overload
from warnings import warn

from pytools import MovedFunctionDeprecationWrapper, memoize


if TYPE_CHECKING or getattr(sys, "_BUILDING_SPHINX_DOCS", None):
    import numpy as np


# {{{ integer powers

def integer_power(x, n, one=1):
    """Compute :math:`x^n` using only multiplications.

    See also the `C2 wiki <https://wiki.c2.com/?IntegerPowerAlgorithm>`__.
    """

    assert isinstance(n, int)

    if n < 0:
        raise RuntimeError("the integer power algorithm does not "
                "work for negative numbers")

    aux = one

    while n > 0:
        if n & 1:
            aux *= x
            if n == 1:
                return aux
        x = x * x
        n //= 2

    return aux

# }}}


# {{{ euclidean algorithm

def extended_euclidean(q, r):
    """Return a tuple *(p, a, b)* such that :math:`p = aq + br`,
    where *p* is the greatest common divisor of *q* and *r*.

    See also the
    `Wikipedia article on the Euclidean algorithm
    <https://en.wikipedia.org/wiki/Euclidean_algorithm>`__.
    """
    import pymbolic.traits as traits

    # see [Davenport], Appendix, p. 214

    t = traits.common_traits(q, r)

    if t.norm(q) < t.norm(r):
        p, a, b = extended_euclidean(r, q)
        return p, b, a

    Q = 1, 0  # noqa
    R = 0, 1  # noqa

    while r:
        quot, t = divmod(q, r)
        T = Q[0] - quot*R[0], Q[1] - quot*R[1]  # noqa
        q, r = r, t
        Q, R = R, T  # noqa: N806

    return q, Q[0], Q[1]


def gcd(q, r):
    return extended_euclidean(q, r)[0]


def gcd_many(*args):
    if len(args) == 0:
        return 1
    elif len(args) == 1:
        return args[0]
    else:
        from functools import reduce
        return reduce(gcd, args)


def lcm(q, r):
    return abs(q*r)//gcd(q, r)

# }}}


# {{{ fft

@memoize
def find_factors(n):
    from math import sqrt

    n1 = 2
    max_n1 = int(sqrt(n))+1
    while n % n1 != 0 and n1 <= max_n1:
        n1 += 1

    if n1 > max_n1:
        n1 = n

    n2 = n // n1

    return n1, n2


def fft(x, sign=1,
        wrap_intermediate=None,
        *,
        wrap_intermediate_with_level=None,
        complex_dtype=None,
        custom_np=None, level=0):
    r"""Computes the Fourier transform of x:

    .. math::

        F[x]_k = \sum_{j=0}^{n-1} z^{kj} x_j

    where :math:`z = \exp(-2i\pi\operatorname{sign}/n)` and ``n == len(x)``.
    Works for all positive *n*.

    See also `Wikipedia <https://en.wikipedia.org/wiki/Cooley%E2%80%93Tukey_FFT_algorithm>`__.
    """

    # revision 293076305
    # https://en.wikipedia.org/w/index.php?title=Cooley-Tukey_FFT_algorithm&oldid=293076305

    # {{{ parameter processing

    if wrap_intermediate is not None and wrap_intermediate_with_level is not None:
        raise TypeError("may specify at most one of wrap_intermediate and "
                "wrap_intermediate_with_level")
    if wrap_intermediate is not None:
        from warnings import warn
        warn("wrap_intermediate is deprecated. Use wrap_intermediate_with_level "
                "instead. wrap_intermediate will stop working in 2023.",
                DeprecationWarning, stacklevel=2)

        def wrap_intermediate_with_level(level, x):  # pylint: disable=function-redefined
            return wrap_intermediate(x)

    if wrap_intermediate_with_level is None:
        def wrap_intermediate_with_level(level, x):
            return x

    from math import pi
    if custom_np is None:
        import numpy as custom_np

    if complex_dtype is None:
        if x.dtype.kind == "c":
            complex_dtype = x.dtype
        else:
            from warnings import warn
            warn("Not supplying complex_dtype is deprecated, falling back "
                    "to complex128 for now. This will stop working in 2023.",
                    DeprecationWarning, stacklevel=2)

            complex_dtype = custom_np.complex128

    complex_dtype = custom_np.dtype(complex_dtype)

    # }}}

    n = len(x)

    if n == 1:
        return x

    N1, N2 = find_factors(n)  # noqa: N806

    scalar_tp = complex_dtype.type
    sub_ffts = [
            wrap_intermediate_with_level(level,
                fft(x[n1::N1], sign, wrap_intermediate, custom_np=custom_np,
                    level=level+1, complex_dtype=complex_dtype)
                * custom_np.exp(
                    sign*-2j*pi*n1/(N1*N2)
                    * custom_np.arange(0, N2, dtype=complex_dtype)))
            for n1 in range(N1)]

    return custom_np.concatenate([
        sum(subvec * scalar_tp(custom_np.exp(sign*(-2j)*pi*n1*k1/N1))
            for n1, subvec in enumerate(sub_ffts))
        for k1 in range(N1)
        ], axis=0)


def ifft(x, wrap_intermediate=None,
         *,
         wrap_intermediate_with_level=None,
         complex_dtype=None,
         custom_np=None):
    return (1/len(x))*fft(x, sign=-1, wrap_intermediate=wrap_intermediate,
            wrap_intermediate_with_level=wrap_intermediate_with_level,
            complex_dtype=complex_dtype, custom_np=custom_np)


def sym_fft(x, sign=1):
    """Perform a (symbolic) FFT on the :mod:`numpy` object array x.

    Remove near-zero floating point constants, insert
    :class:`pymbolic.primitives.CommonSubexpression`
    wrappers at opportune points.
    """

    from pymbolic.mapper import CSECachingMapperMixin, IdentityMapper

    class NearZeroKiller(CSECachingMapperMixin, IdentityMapper):
        map_common_subexpression_uncached = \
                IdentityMapper.map_common_subexpression

        def map_constant(self, expr):
            if isinstance(expr, complex):
                r = expr.real
                i = expr.imag
                if abs(r) < 1e-15:
                    r = 0
                if abs(i) < 1e-15:
                    i = 0
                if i == 0:
                    return r
                else:
                    return complex(r, i)
            else:
                return expr

    import numpy

    def wrap_intermediate(x):
        if len(x) > 1:
            from pymbolic.primitives import CommonSubexpression, cse_scope

            result = numpy.empty(len(x), dtype=object)
            for i, x_i in enumerate(x):
                result[i] = CommonSubexpression(x_i, scope=cse_scope.EVALUATION)

            return result
        else:
            return x

    return NearZeroKiller()(
            fft(wrap_intermediate(x), sign=sign,
                wrap_intermediate=wrap_intermediate))

# }}}


def csr_matrix_multiply(S, x):  # noqa
    """Multiplies a :class:`scipy.sparse.csr_matrix` S by an object-array vector x.
    """
    h, _w = S.shape

    import numpy
    result = numpy.empty_like(x)

    for i in range(h):
        result[i] = sum(S.data[idx]*x[S.indices[idx]]  # pylint:disable=unsupported-assignment-operation
                for idx in range(S.indptr[i], S.indptr[i+1]))

    return result


# {{{ reduced_row_echelon_form

@overload
def reduced_row_echelon_form(
            mat: np.ndarray,
            *, integral: bool | None = None,
        ) -> np.ndarray:
    ...


@overload
def reduced_row_echelon_form(
            mat: np.ndarray,
            rhs: np.ndarray,
            *, integral: bool | None = None,
        ) -> tuple[np.ndarray, np.ndarray]:
    ...


def reduced_row_echelon_form(
            mat: np.ndarray,
            rhs: np.ndarray | None = None,
            integral: bool | None = None,
        ) -> tuple[np.ndarray, np.ndarray] | np.ndarray:
    m, n = mat.shape

    mat = mat.copy()
    if rhs is not None:
        rhs = rhs.copy()

    if integral is None:
        warn(
             "Not specifying 'integral' is deprecated, please add it as an argument. "
             "This will stop being supported in 2025.",
             DeprecationWarning, stacklevel=2)

    if integral:
        div_func = operator.floordiv
    else:
        div_func = operator.truediv

    i = 0
    j = 0

    while i < m and j < n:
        # {{{ find pivot in column j, starting in row i

        nonz_row = None
        for k in range(i, m):
            if mat[k, j]:
                nonz_row = k
                break

        # }}}

        if nonz_row is not None:
            # swap rows i and nonz
            mat[i], mat[nonz_row] = \
                    (mat[nonz_row].copy(), mat[i].copy())
            if rhs is not None:
                rhs[i], rhs[nonz_row] = \
                        (rhs[nonz_row].copy(), rhs[i].copy())

            for u in range(0, m):
                if u == i:
                    continue
                if not mat[u, j]:
                    # already 0
                    continue

                ell = lcm(mat[u, j], mat[i, j])
                u_fac = div_func(ell, mat[u, j])
                i_fac = div_func(ell, mat[i, j])

                mat[u] = u_fac*mat[u] - i_fac*mat[i]
                if rhs is not None:
                    rhs[u] = u_fac*rhs[u] - i_fac*rhs[i]

                assert mat[u, j] == 0

            i += 1

        j += 1

    if integral:
        for i in range(m):
            g = gcd_many(*(
                [a for a in mat[i] if a]
                +
                [a for a in rhs[i] if a] if rhs is not None else []))

            mat[i] //= g
            if rhs is not None:
                rhs[i] //= g

    import numpy as np

    from pymbolic.mapper.flattener import flatten
    vec_flatten = np.vectorize(flatten, otypes=[object])

    for i in range(m):
        mat[i] = vec_flatten(mat[i])
        if rhs is not None:
            rhs[i] = vec_flatten(rhs[i])

    if rhs is None:
        return mat
    else:
        return mat, rhs

# }}}


gaussian_elimination = MovedFunctionDeprecationWrapper(reduced_row_echelon_form, "2025")


# {{{ symbolic (linear) equation solving

def solve_affine_equations_for(unknowns, equations):
    """
    :arg unknowns: A list of variable names for which to solve.
    :arg equations: A list of tuples ``(lhs, rhs)``.
    :return: a dict mapping unknown names to their values.
    """
    import numpy as np

    from pymbolic.mapper.dependency import DependencyMapper
    dep_map = DependencyMapper(composite_leaves=True)

    # fix an order for unknowns
    from pymbolic import var
    unknowns = [var(u) for u in unknowns]
    unknowns_set = set(unknowns)
    unknown_idx_lut = {tgt_name: idx
            for idx, tgt_name in enumerate(unknowns)}

    # Find non-unknown variables, fix order for them
    # Last non-unknown is constant.
    parameters = set()
    for lhs, rhs in equations:
        parameters.update(dep_map(lhs) - unknowns_set)
        parameters.update(dep_map(rhs) - unknowns_set)

    parameters_list = list(parameters)
    parameter_idx_lut = {var_name: idx
            for idx, var_name in enumerate(parameters_list)}

    from pymbolic.mapper.coefficient import CoefficientCollector
    coeff_coll = CoefficientCollector()

    # {{{ build matrix and rhs

    mat = np.zeros((len(equations), len(unknowns_set)), dtype=object)
    rhs_mat = np.zeros((len(equations), len(parameters)+1), dtype=object)

    for i_eqn, (lhs, rhs) in enumerate(equations):
        for lhs_factor, coeffs in [(1, coeff_coll(lhs)), (-1, coeff_coll(rhs))]:
            for key, coeff in coeffs.items():
                if key in unknowns_set:
                    mat[i_eqn, unknown_idx_lut[key]] = lhs_factor*coeff
                elif key in parameters:
                    rhs_mat[i_eqn, parameter_idx_lut[key]] = -lhs_factor*coeff
                elif key == 1:
                    rhs_mat[i_eqn, -1] = -lhs_factor*coeff
                else:
                    raise ValueError(f"key '{key}' not understood")

    # }}}

    mat, rhs_mat = reduced_row_echelon_form(mat, rhs_mat, integral=True)

    # FIXME /!\ Does not check for overdetermined system.

    result = {}
    for j, unknown in enumerate(unknowns):
        (nonz_row,) = np.where(mat[:, j])
        if len(nonz_row) != 1:
            raise RuntimeError(f"cannot uniquely solve for '{unknown}'")

        (nonz_row,) = nonz_row

        if abs(mat[nonz_row, j]) != 1:
            raise RuntimeError(
                    f"division with remainder in linear solve for '{unknown}'")
        div = mat[nonz_row, j]

        unknown_val = int(rhs_mat[nonz_row, -1]) // div
        for parameter, coeff in zip(
                    parameters_list, rhs_mat[nonz_row, :-1], strict=True):
            unknown_val += (int(coeff) // div) * parameter

        result[unknown] = unknown_val

    if 0:
        for lhs, rhs in equations:
            print(lhs, "=", rhs)
        print("-------------------")
        for lhs, rhs in result.items():
            print(lhs, "=", rhs)

    return result

# }}}


# vim: foldmethod=marker
